# Copyright (C) 2024 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains functions for making web requests with Toil.

All web requests should go through this module, to make sure they use the right
user agent.

>>> from toil.lib.web import web_session
>>> web_session.get("https://example.com")

"""

import logging
import requests
import sys

from toil.version import baseVersion

# We manage a Requests session at the module level in case we're supposed to be
# doing cookies, and to send a sensible user agent.
# We expect the Toil and Python version to not be personally identifiable even
# in theory (someone might make a new Toil version first, but there's no way
# to know for sure that nobody else did the same thing).
web_session = requests.Session()
web_session.headers.update({"User-Agent": f"Toil {baseVersion} on Python {'.'.join([str(v) for v in sys.version_info])}"})



#!/usr/bin/env python3
# example_cachingbenchmark.py: Empirically inspect Toil caching

"""
This workflow collects statistics about caching.

Invoke like:

    python examples/example_cachingbenchmark.py ./jobstore --realTimeLogging --logInfo

    python examples/example_cachingbenchmark.py ./jobstore --realTimeLogging --logInfo --disableCaching

    python examples/example_cachingbenchmark.py aws:us-west-2:cachingjobstore --realTimeLogging --logInfo

    python examples/example_cachingbenchmark.py aws:us-west-2:cachingjobstore --realTimeLogging --logInfo --disableCaching

"""

import argparse
import collections
import os
import random
import socket
import sys
import time

from configargparse import ArgumentParser

from toil.common import Toil
from toil.job import Job
from toil.realtimeLogger import RealtimeLogger


def main():
    parser = ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument(
        "--minSleep", type=int, default=1, help="Minimum seconds to sleep"
    )

    Job.Runner.addToilOptions(parser)

    options = parser.parse_args(sys.argv[1:])

    root_job = Job.wrapJobFn(root, options)

    with Toil(options) as toil:
        results = toil.start(root_job)

    print("Caching results:")
    print(results)


def root(job, options):
    # Make a file
    with job.fileStore.writeGlobalFileStream() as (stream, file_id):
        stream.write(
            b"This is a test of the Toil file caching system. "
            b"Had this been an actual file, its contents would have been more interesting."
        )

    child_rvs = []
    for i in range(100):
        # Make lots of child jobs that read and report on the file
        child_rvs.append(job.addChildJobFn(poll, options, file_id, i).rv())

    # Collect all their views into a report
    return job.addFollowOnJobFn(report, child_rvs).rv()


def poll(job, options, file_id, number, cores=0.1, disk="200M", memory="512M"):

    # Wait a random amount of time before grabbing the file for others to cache it
    time.sleep(random.randint(options.minSleep, options.minSleep + 10))

    # Read the file. Don't accept a symlink because then we might just have the
    # filestore's copy, even if caching is not happening.
    local_file = job.fileStore.readGlobalFile(
        file_id, cache=True, mutable=False, symlink=False
    )

    # Wait a random amount of after before grabbing the file for others to use it
    time.sleep(random.randint(options.minSleep, options.minSleep + 10))

    # Stat the file (reads through links)
    stats = os.stat(local_file)

    # Check what machine we are
    hostname = socket.gethostname()

    RealtimeLogger.info(
        f"Job {number} on host {hostname} sees file at device {stats.st_dev} inode {stats.st_ino}"
    )

    # Return a tuple representing our view of the file.
    # Drop hostname since hostnames are unique per pod.
    return stats.st_dev, stats.st_ino


def report(job, views):
    # Count the distinct views
    counts = collections.Counter()
    for v in views:
        counts[v] += 1

    report = [f"{len(counts)} distinct views, most frequent:"]

    for view, count in counts.most_common(10):
        report.append(f"{view}: {count}")

    return "\n".join(report)


if __name__ == "__main__":
    main()



"""lsfHelper.py shouldn't need a batch system and so the unit tests here should aim to run on any system."""

from toil.batchSystems.lsfHelper import parse_mem_and_cmd_from_output
from toil.test import ToilTest


class LSFHelperTest(ToilTest):
    def test_parse_mem_and_cmd_from_output(self):
        # https://github.com/DataBiosphere/toil/pull/3475
        output = (
            "\nJob <2924748>, Job Name <toil_job_64>, User <thiagogenez>, Project <default>, S"
            "\n                     tatus <RUN>, Queue <research-rh74>, Job Priority <50>, Com"
            "\n                     mand <_toil_worker CactusBarRecursion file:/hps/nobackup/p"
            "\n                     roduction/ensembl/thiagogenez/pairwises/arabidopsis/run/jo"
            "\n                     bstore/3 kind-CactusBarRecursion/instance-iu6wo56x --conte"
            "\n                     xt gAShortenedh32xqlE51Yi4=>, Share group charged </thiago"
            "\n                     genez>, Esub <esub>"
            "\nThu Mar 18 02:06:32: Submitted from host <noah-login-01>, CWD </hps/nobackup/pr"
            "\n                     oduction/ensembl/thiagogenez/pairwises/arabidopsis/run>, S"
            "\n                     pecified CWD </hps/nobackup/production/ensembl/thiagogenez"
            "\n                     /pairwises/arabidopsis/run/.>, Output File </tmp/toil_work"
            "\n                     flow_10e83102-2e4b-4093-9128-2a52f4bbc65a_job_64_batch_lsf"
            "\n                     _2924748_std_output.log>, Error File </tmp/toil_workflow_1"
            "\n                     0e83102-2e4b-4093-9128-2a52f4bbc65a_job_64_batch_lsf_29247"
            "\n                     48_std_error.log>, Requested Resources <select[(mem>4000)]"
            "\n                      rusage[mem=4000:duration=480, numcpus=1:duration=480]>;"
            "\nThu Mar 18 02:06:33: Started on <hx-noah-31-02>, Execution Home </homes/thiagog"
            "\n                     enez>, Execution CWD </hps/nobackup/production/ensembl/thi"
            "\n                     agogenez/pairwises/arabidopsis/run/.>;"
            "\nThu Mar 18 17:07:47: Resource usage collected."
            "\n                     The CPU time used is 53936 seconds."
            "\n                     MEM: 344 Mbytes;  SWAP: 1.3 Gbytes;  NTHREAD: 5"
            "\n                     PGID: 433168;  PIDs: 433168 433177 433179 444026 "
            "\n"
            "\n RUNLIMIT                "
            "\n 10085.0 min"
            "\n"
            "\n CORELIMIT MEMLIMIT"
            "\n      0 M     3.9 G "
            "\n"
            "\n MEMORY USAGE:"
            "\n MAX MEM: 2.5 Gbytes;  AVG MEM: 343 Mbytes"
            "\n"
            "\n SCHEDULING PARAMETERS:"
            "\n           r15s   r1m  r15m   ut      pg    io   ls    it    tmp    swp    mem"
            "\n loadSched   -     -     -     -    10.0     -    -     -   500M     -   1000M "
            "\n loadStop    -     -     -     -       -     -    -     -     -      -      -  "
            "\n"
            "\n          availcpus "
            "\n loadSched      1.0 "
            "\n loadStop        -  "
            "\n"
            "\n RESOURCE REQUIREMENT DETAILS:"
            "\n Combined: select[((mem>4000)) && (type == local)] order[r15s:pg] rusage[mem=40"
            "\n                     00.00:duration=8h:decay=0,numcpus=1.00:duration=8h:decay=0"
            "\n                     ] span[hosts=1]"
            "\n Effective: select[(((mem>4000))) && (type == local)] order[r15s:pg] rusage[mem"
            "\n                     =4000.00:duration=8h:decay=0,numcpus=1.00:duration=8h:deca"
            "\n                     y=0] span[hosts=1] "
            "\n"
            "\n"
        )
        max_mem, command = parse_mem_and_cmd_from_output(output=output)
        assert len(max_mem.groups()) == 1
        expected_mem = "2.5 Gbytes"
        assert (
            max_mem.group(1) == expected_mem
        ), f'Actual: {max_mem.group(1)}, Expected: "{expected_mem}"'

        assert len(command.groups()) == 1
        expected_command = (
            "_toil_worker CactusBarRecursion file:/hps/nobackup/production/ensembl/thiagogenez/"
            "pairwises/arabidopsis/run/jobstore/3 kind-CactusBarRecursion/instance-iu6wo56x "
            "--context gAShortenedh32xqlE51Yi4="
        )
        assert (
            command.group(1) == expected_command
        ), f'Actual: {command.group(1)}, Expected: "{expected_command}"'
        print(command)

        max_mem, command = parse_mem_and_cmd_from_output(output="")
        assert max_mem == None
        assert command == None



import logging
import subprocess
import time
from contextlib import contextmanager

from toil.exceptions import FailedJobsException
from toil.lib.iterables import concat
from toil.test import ApplianceTestSupport, needs_local_appliance, needs_mesos, slow
from toil.version import exactPython

logger = logging.getLogger(__name__)


@needs_mesos
@needs_local_appliance
@slow
class AutoDeploymentTest(ApplianceTestSupport):
    """
    Tests various auto-deployment scenarios. Using the appliance, i.e. a docker container,
    for these tests allows for running worker processes on the same node as the leader process
    while keeping their file systems separate from each other and the leader process. Separate
    file systems are crucial to prove that auto-deployment does its job.
    """

    def setUp(self):
        logging.basicConfig(level=logging.INFO)
        super().setUp()

    @contextmanager
    def _venvApplianceCluster(self):
        """
        Creates an appliance cluster with a virtualenv at './venv' on the leader and a temporary
        directory on the host mounted at /data in the leader and worker containers.
        """
        dataDirPath = self._createTempDir(purpose="data")
        with self._applianceCluster(mounts={dataDirPath: "/data"}) as (leader, worker):
            leader.runOnAppliance(
                "virtualenv",
                "--system-site-packages",
                "--never-download",  # prevent silent upgrades to pip etc
                "--python",
                exactPython,
                "venv",
            )
            leader.runOnAppliance("venv/bin/pip", "list")  # For diagnostic purposes
            yield leader, worker

    # TODO: Are we sure the python in the appliance we are testing is the same
    # as the one we are testing from? If not, how can we get the version it is?
    sitePackages = f"venv/lib/{exactPython}/site-packages"

    def testRestart(self):
        """
        Test whether auto-deployment works on restart.
        """
        with self._venvApplianceCluster() as (leader, worker):

            def userScript():
                from toil.common import Toil
                from toil.job import Job

                # noinspection PyUnusedLocal
                def job(job, disk="10M", cores=1, memory="10M"):
                    assert False

                if __name__ == "__main__":
                    options = Job.Runner.getDefaultArgumentParser().parse_args()
                    with Toil(options) as toil:
                        if toil.config.restart:
                            toil.restart()
                        else:
                            toil.start(Job.wrapJobFn(job))

            userScript = self._getScriptSource(userScript)

            leader.deployScript(
                path=self.sitePackages, packagePath="foo.bar", script=userScript
            )

            pythonArgs = ["venv/bin/python", "-m", "foo.bar"]
            toilArgs = [
                "--logDebug",
                "--batchSystem=mesos",
                "--mesosEndpoint=localhost:5050",
                "--defaultMemory=10M",
                "/data/jobstore",
            ]
            command = concat(pythonArgs, toilArgs)
            self.assertRaises(
                subprocess.CalledProcessError, leader.runOnAppliance, *command
            )

            # Deploy an updated version of the script ...
            userScript = userScript.replace("assert False", "assert True")
            leader.deployScript(
                path=self.sitePackages, packagePath="foo.bar", script=userScript
            )
            # ... and restart Toil.
            command = concat(pythonArgs, "--restart", toilArgs)
            leader.runOnAppliance(*command)

    def testSplitRootPackages(self):
        """
        Test whether auto-deployment works with a virtualenv in which jobs are defined in
        completely separate branches of the package hierarchy. Initially, auto-deployment did
        deploy the entire virtualenv but jobs could only be defined in one branch of the package
        hierarchy. We define a branch as the maximum set of fully qualified package paths that
        share the same first component. IOW, a.b and a.c are in the same branch, while a.b and
        d.c are not.
        """
        with self._venvApplianceCluster() as (leader, worker):

            # Deploy the library module with job definitions
            def libraryModule():
                # noinspection PyUnusedLocal
                def libraryJob(job):
                    open("/data/foo.txt", "w").close()

            leader.deployScript(
                path=self.sitePackages, packagePath="toil_lib.foo", script=libraryModule
            )

            # Deploy the user script
            def userScript():
                # noinspection PyUnresolvedReferences
                from toil_lib.foo import libraryJob

                from toil.common import Toil
                from toil.job import Job

                # noinspection PyUnusedLocal
                def job(job, disk="10M", cores=1, memory="10M"):
                    # Double the requirements to prevent chaining as chaining might hide problems
                    # in auto-deployment code.
                    job.addChildJobFn(
                        libraryJob, disk="20M", cores=cores, memory=memory
                    )

                if __name__ == "__main__":
                    options = Job.Runner.getDefaultArgumentParser().parse_args()
                    with Toil(options) as toil:
                        if toil.config.restart:
                            toil.restart()
                        else:
                            toil.start(Job.wrapJobFn(job))

            leader.deployScript(
                path=self.sitePackages, packagePath="toil_script.bar", script=userScript
            )

            # Assert that output file isn't there
            worker.runOnAppliance("test", "!", "-f", "/data/foo.txt")
            # Just being paranoid
            self.assertRaises(
                subprocess.CalledProcessError,
                worker.runOnAppliance,
                "test",
                "-f",
                "/data/foo.txt",
            )
            leader.runOnAppliance(
                "venv/bin/python",
                "-m",
                "toil_script.bar",
                "--logDebug",
                "--batchSystem=mesos",
                "--mesosEndpoint=localhost:5050",
                "--defaultMemory=10M",
                "/data/jobstore",
            )
            # Assert that out output file is there
            worker.runOnAppliance("test", "-f", "/data/foo.txt")

    def testUserTypesInJobFunctionArgs(self):
        """
        Test encapsulated, function-wrapping jobs where the function arguments reference
        user-defined types.

        Mainly written to cover https://github.com/BD2KGenomics/toil/issues/1259 but then also
        revealed https://github.com/BD2KGenomics/toil/issues/1278.
        """
        with self._venvApplianceCluster() as (leader, worker):

            def userScript():
                from toil.common import Toil
                from toil.job import Job

                # A user-defined type, i.e. a type defined in the user script
                class X:
                    pass

                # noinspection PyUnusedLocal
                def job(job, x, disk="10M", cores=1, memory="10M"):
                    return x

                if __name__ == "__main__":
                    options = Job.Runner.getDefaultArgumentParser().parse_args()
                    x = X()
                    with Toil(options) as toil:
                        r = toil.start(Job.wrapJobFn(job, x).encapsulate())
                    # Assert that the return value is of type X, but not X from the __main__
                    # module but X from foo.bar, the canonical name for the user module. The
                    # translation from __main__ to foo.bar is a side effect of auto-deployment.
                    assert r.__class__ is not X
                    import foo.bar

                    assert r.__class__ is foo.bar.X
                    # Assert that a copy was made. This is a side effect of pickling/unpickling.
                    assert x is not r

            userScript = self._getScriptSource(userScript)

            leader.deployScript(
                path=self.sitePackages, packagePath="foo.bar", script=userScript
            )

            leader.runOnAppliance(
                "venv/bin/python",
                "-m",
                "foo.bar",
                "--logDebug",
                "--batchSystem=mesos",
                "--mesosEndpoint=localhost:5050",
                "--defaultMemory=10M",
                "--defaultDisk=10M",
                "/data/jobstore",
            )

    def testDeferralWithConcurrentEncapsulation(self):
        """
        Ensure that the following DAG succeeds::

                          ┌───────────┐
                          │ Root (W1) │
                          └───────────┘
                                │
                     ┌──────────┴─────────┐
                     ▼                    ▼
            ┌────────────────┐ ┌────────────────────┐
            │ Deferring (W2) │ │ Encapsulating (W3) │═══════════════╗
            └────────────────┘ └────────────────────┘               ║
                                          │                         ║
                                          ▼                         ▼
                                ┌───────────────────┐      ┌────────────────┐
                                │ Encapsulated (W3) │      │ Follow-on (W6) │
                                └───────────────────┘      └────────────────┘
                                          │                         │
                                  ┌───────┴────────┐                │
                                  ▼                ▼                ▼
                          ┌──────────────┐ ┌──────────────┐ ┌──────────────┐
                          │ Dummy 1 (W4) │ │ Dummy 2 (W5) │ │  Last (W6)   │
                          └──────────────┘ └──────────────┘ └──────────────┘

        The Wn numbers denote the worker processes that a particular job is run in. `Deferring`
        adds a deferred function and then runs for a long time. The deferred function will be
        present in the cache state for the duration of `Deferred`. `Follow-on` is the generic Job
        instance that's added by encapsulating a job. It runs on the same worker node but in a
        separate worker process, as the first job in that worker. Because …

        1) it is the first job in its worker process (the user script has not been made available
        on the sys.path by a previous job in that worker) and

        2) it shares the cache state with the `Deferring` job and

        3) it is an instance of Job (and so does not introduce the user script to sys.path itself),

        … it might cause problems with deserializing a deferred function defined in the user script.

        `Encapsulated` has two children to ensure that `Follow-on` is run in a separate worker.
        """
        with self._venvApplianceCluster() as (leader, worker):

            def userScript():
                from toil.common import Toil
                from toil.job import Job

                def root(rootJob):
                    def nullFile():
                        return rootJob.fileStore.jobStore.import_file(
                            "file:///dev/null"
                        )

                    startFile = nullFile()
                    endFile = nullFile()
                    rootJob.addChildJobFn(deferring, startFile, endFile)
                    encapsulatedJob = Job.wrapJobFn(encapsulated, startFile)
                    encapsulatedJob.addChildFn(dummy)
                    encapsulatedJob.addChildFn(dummy)
                    encapsulatingJob = encapsulatedJob.encapsulate()
                    rootJob.addChild(encapsulatingJob)
                    encapsulatingJob.addChildJobFn(last, endFile)

                def dummy():
                    pass

                def deferred():
                    pass

                # noinspection PyUnusedLocal
                def deferring(job, startFile, endFile):
                    job.defer(deferred)
                    job.fileStore.jobStore.delete_file(startFile)
                    timeout = time.time() + 10
                    while job.fileStore.jobStore.file_exists(endFile):
                        assert time.time() < timeout
                        time.sleep(1)

                def encapsulated(job, startFile):
                    timeout = time.time() + 10
                    while job.fileStore.jobStore.file_exists(startFile):
                        assert time.time() < timeout
                        time.sleep(1)

                def last(job, endFile):
                    job.fileStore.jobStore.delete_file(endFile)

                if __name__ == "__main__":
                    options = Job.Runner.getDefaultArgumentParser().parse_args()
                    with Toil(options) as toil:
                        rootJob = Job.wrapJobFn(root)
                        toil.start(rootJob)

            userScript = self._getScriptSource(userScript)

            leader.deployScript(
                path=self.sitePackages, packagePath="foo.bar", script=userScript
            )

            leader.runOnAppliance(
                "venv/bin/python",
                "-m",
                "foo.bar",
                "--logDebug",
                "--batchSystem=mesos",
                "--mesosEndpoint=localhost:5050",
                "--retryCount=0",
                "--defaultMemory=10M",
                "--defaultDisk=10M",
                "/data/jobstore",
            )

    def testDeferralWithFailureAndEncapsulation(self):
        """
        Ensure that the following DAG succeeds::

                          ┌───────────┐
                          │ Root (W1) │
                          └───────────┘
                                │
                     ┌──────────┴─────────┐
                     ▼                    ▼
            ┌────────────────┐ ┌────────────────────┐
            │ Deferring (W2) │ │ Encapsulating (W3) │═══════════════════════╗
            └────────────────┘ └────────────────────┘                       ║
                                          │                                 ║
                                          ▼                                 ▼
                                ┌───────────────────┐              ┌────────────────┐
                                │ Encapsulated (W3) │════════════╗ │ Follow-on (W7) │
                                └───────────────────┘            ║ └────────────────┘
                                          │                      ║
                                   ┌──────┴──────┐               ║
                                   ▼             ▼               ▼
                            ┌────────────┐┌────────────┐ ┌──────────────┐
                            │ Dummy (W4) ││ Dummy (W5) │ │ Trigger (W6) │
                            └────────────┘└────────────┘ └──────────────┘

        `Trigger` causes `Deferring` to crash. `Follow-on` runs next, detects `Deferring`'s
        left-overs and runs the deferred function. `Follow-on` is an instance of `Job` and the
        first job in its worker process. This test ensures that despite these circumstances,
        the user script is loaded before the deferred functions defined in it are being run.

        `Encapsulated` has two children to ensure that `Follow-on` is run in a new worker. That's
        the only way to guarantee that the user script has not been loaded yet, which would cause
        the test to succeed coincidentally. We want to test that auto-deploying and loading of the
        user script are done properly *before* deferred functions are being run and before any
        jobs have been executed by that worker.
        """
        with self._venvApplianceCluster() as (leader, worker):

            def userScript():
                import os
                import time

                from toil.common import Toil
                from toil.job import Job

                TIMEOUT = 10

                def root(rootJob):
                    def nullFile():
                        return rootJob.fileStore.jobStore.import_file(
                            "file:///dev/null"
                        )

                    startFile = nullFile()
                    endFile = nullFile()

                    rootJob.addChildJobFn(deferring, startFile, endFile)
                    encapsulatedJob = Job.wrapJobFn(encapsulated, startFile)
                    encapsulatedJob.addChildFn(dummy)
                    encapsulatedJob.addChildFn(dummy)
                    encapsulatedJob.addFollowOnJobFn(trigger, endFile)
                    encapsulatingJob = encapsulatedJob.encapsulate()
                    rootJob.addChild(encapsulatingJob)

                def dummy():
                    pass

                def deferredFile(config):
                    """
                    Return path to a file at the root of the job store, exploiting the fact that
                    the job store is shared between leader and worker container.
                    """
                    prefix = "file:"
                    locator = config.jobStore
                    assert locator.startswith(prefix)
                    return os.path.join(locator[len(prefix) :], "testDeferredFile")

                def deferred(deferredFilePath):
                    """
                    The deferred function that is supposed to run.
                    """
                    os.unlink(deferredFilePath)

                # noinspection PyUnusedLocal
                def deferring(job, startFile, endFile):
                    """
                    A job that adds the deferred function and then crashes once the `trigger` job
                    tells it to.
                    """
                    job.defer(deferred, deferredFile(job._config))
                    jobStore = job.fileStore.jobStore
                    jobStore.delete_file(startFile)
                    with jobStore.update_file_stream(endFile) as fH:
                        fH.write(str(os.getpid()))
                    timeout = time.time() + TIMEOUT
                    while jobStore.file_exists(endFile):
                        assert time.time() < timeout
                        time.sleep(1)
                    os.kill(os.getpid(), 9)

                def encapsulated(job, startFile):
                    """
                    A job that waits until the `deferring` job is running and waiting to be crashed.
                    """
                    timeout = time.time() + TIMEOUT
                    while job.fileStore.jobStore.file_exists(startFile):
                        assert time.time() < timeout
                        time.sleep(1)

                def trigger(job, endFile):
                    """
                    A job that determines the PID of the worker running the `deferring` job,
                    tells the `deferring` job to crash and then waits for the corresponding
                    worker process to end. By waiting we can be sure that the `follow-on` job
                    finds the left-overs of the `deferring` job.
                    """
                    import errno

                    jobStore = job.fileStore.jobStore
                    with jobStore.read_file_stream(endFile) as fH:
                        pid = int(fH.read())
                    os.kill(pid, 0)
                    jobStore.delete_file(endFile)
                    timeout = time.time() + TIMEOUT
                    while True:
                        try:
                            os.kill(pid, 0)
                        except OSError as e:
                            if e.errno == errno.ESRCH:
                                break
                            else:
                                raise
                        else:
                            assert time.time() < timeout
                            time.sleep(1)

                def tryUnlink(deferredFilePath):
                    try:
                        os.unlink(deferredFilePath)
                    except OSError as e:
                        if e.errno == errno.ENOENT:
                            pass
                        else:
                            raise

                if __name__ == "__main__":
                    import errno

                    options = Job.Runner.getDefaultArgumentParser().parse_args()
                    with Toil(options) as toil:
                        deferredFilePath = deferredFile(toil.config)
                        open(deferredFilePath, "w").close()
                        try:
                            assert os.path.exists(deferredFilePath)
                            try:
                                toil.start(Job.wrapJobFn(root))
                            except FailedJobsException as e:
                                assert (
                                    e.numberOfFailedJobs == 2
                                )  # `root` and `deferring`
                                assert not os.path.exists(
                                    deferredFilePath
                                ), "Apparently, the deferred function did not run."
                            else:
                                assert False, "Workflow should not have succeeded."
                        finally:
                            tryUnlink(deferredFilePath)

            userScript = self._getScriptSource(userScript)

            leader.deployScript(
                path=self.sitePackages, packagePath="foo.bar", script=userScript
            )

            leader.runOnAppliance(
                "venv/bin/python",
                "-m",
                "foo.bar",
                "--logDebug",
                "--batchSystem=mesos",
                "--mesosEndpoint=localhost:5050",
                "--retryCount=0",
                "--defaultMemory=10M",
                "--defaultDisk=10M",
                "/data/jobstore",
            )



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import subprocess
import time
from typing import Optional
from uuid import uuid4

from toil.lib.aws import zone_to_region
from toil.lib.aws.session import AWSConnectionManager
from toil.lib.retry import retry
from toil.provisioners import cluster_factory
from toil.provisioners.aws import get_best_aws_zone
from toil.test import (
    ToilTest,
    needs_aws_ec2,
    needs_env_var,
    needs_fetchable_appliance,
    slow,
)

log = logging.getLogger(__name__)


@needs_aws_ec2
@needs_fetchable_appliance
class AbstractClusterTest(ToilTest):
    def __init__(self, methodName: str) -> None:
        super().__init__(methodName=methodName)
        self.keyName = os.getenv("TOIL_AWS_KEYNAME", "id_rsa").strip()
        self.clusterName = f"aws-provisioner-test-{uuid4()}"
        self.leaderNodeType = "t2.medium"
        self.clusterType = "mesos"
        self.zone = get_best_aws_zone()
        assert (
            self.zone is not None
        ), "Could not determine AWS availability zone to test in; is TOIL_AWS_ZONE set?"
        self.region = zone_to_region(self.zone)

        # Get connection to AWS
        self.aws = AWSConnectionManager()

        # Where should we put our virtualenv?
        self.venvDir = "/tmp/venv"

    def python(self) -> str:
        """
        Return the full path to the venv Python on the leader.
        """
        return os.path.join(self.venvDir, "bin/python")

    def pip(self) -> str:
        """
        Return the full path to the venv pip on the leader.
        """
        return os.path.join(self.venvDir, "bin/pip")

    def destroyCluster(self) -> None:
        """
        Destroy the cluster we built, if it exists.

        Succeeds if the cluster does not currently exist.
        """
        subprocess.check_call(
            ["toil", "destroy-cluster", "-p=aws", "-z", self.zone, self.clusterName]
        )

    def setUp(self) -> None:
        """
        Set up for the test.
        Must be overridden to call this method and set self.jobStore.
        """
        super().setUp()
        # Make sure that destroy works before we create any clusters.
        # If this fails, no tests will run.
        self.destroyCluster()

    def tearDown(self) -> None:
        # Note that teardown will run even if the test crashes.
        super().tearDown()
        self.destroyCluster()
        subprocess.check_call(["toil", "clean", self.jobStore])

    def sshUtil(self, command: list[str]) -> None:
        """
        Run the given command on the cluster.
        Raise subprocess.CalledProcessError if it fails.
        """

        cmd = [
            "toil",
            "ssh-cluster",
            "--insecure",
            "-p=aws",
            "-z",
            self.zone,
            self.clusterName,
        ] + command
        log.info("Running %s.", str(cmd))
        p = subprocess.Popen(cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
        # Put in non-blocking mode. See https://stackoverflow.com/a/59291466
        os.set_blocking(p.stdout.fileno(), False)
        os.set_blocking(p.stderr.fileno(), False)

        out_buffer = b""
        err_buffer = b""

        loops_since_line = 0

        running = True
        while running:
            # While the process is running, see if it stopped
            running = p.poll() is None

            # Also collect its output
            out_data = p.stdout.read()
            if out_data:
                out_buffer += out_data

            while out_buffer.find(b"\n") != -1:
                # And log every full line
                cut = out_buffer.find(b"\n")
                log.info(
                    "STDOUT: %s", out_buffer[0:cut].decode("utf-8", errors="ignore")
                )
                loops_since_line = 0
                out_buffer = out_buffer[cut + 1 :]

            # Same for the error
            err_data = p.stderr.read()
            if err_data:
                err_buffer += err_data

            while err_buffer.find(b"\n") != -1:
                cut = err_buffer.find(b"\n")
                log.info(
                    "STDERR: %s", err_buffer[0:cut].decode("utf-8", errors="ignore")
                )
                loops_since_line = 0
                err_buffer = err_buffer[cut + 1 :]

            loops_since_line += 1
            if loops_since_line > 60:
                log.debug("...waiting...")
                loops_since_line = 0

            time.sleep(1)

        # At the end, log the last lines
        if out_buffer:
            log.info("STDOUT: %s", out_buffer.decode("utf-8", errors="ignore"))
        if err_buffer:
            log.info("STDERR: %s", err_buffer.decode("utf-8", errors="ignore"))

        if p.returncode != 0:
            # It failed
            log.error("Failed to run %s.", str(cmd))
            raise subprocess.CalledProcessError(p.returncode, " ".join(cmd))

    @retry(errors=[subprocess.CalledProcessError], intervals=[1, 1])
    def rsync_util(self, from_file: str, to_file: str) -> None:
        """
        Transfer a file to/from the cluster.

        The cluster-side path should have a ':' in front of it.
        """
        cmd = [
            "toil",
            "rsync-cluster",
            "--insecure",
            "-p=aws",
            "-z",
            self.zone,
            self.clusterName,
            from_file,
            to_file,
        ]
        log.info("Running %s.", str(cmd))
        subprocess.check_call(cmd)

    @retry(errors=[subprocess.CalledProcessError], intervals=[1, 1])
    def createClusterUtil(self, args: Optional[list[str]] = None) -> None:
        args = [] if args is None else args

        command = [
            "toil",
            "launch-cluster",
            "-p=aws",
            "-z",
            self.zone,
            f"--keyPairName={self.keyName}",
            f"--leaderNodeType={self.leaderNodeType}",
            f"--clusterType={self.clusterType}",
            "--logDebug",
            self.clusterName,
        ] + args

        log.debug("Launching cluster: %s", command)

        # Try creating the cluster
        subprocess.check_call(command)
        # If we fail, tearDown will destroy the cluster.

    def launchCluster(self) -> None:
        self.createClusterUtil()


@needs_aws_ec2
@needs_fetchable_appliance
@slow
class CWLOnARMTest(AbstractClusterTest):
    """Run the CWL 1.2 conformance tests on ARM specifically."""

    def __init__(self, methodName: str) -> None:
        super().__init__(methodName=methodName)
        self.clusterName = f"cwl-test-{uuid4()}"
        self.leaderNodeType = "t4g.2xlarge"
        self.clusterType = "kubernetes"
        # We need to be running in a directory which Flatcar and the Toil Appliance both have
        self.cwl_test_dir = "/tmp/toil/cwlTests"

    def setUp(self) -> None:
        super().setUp()
        self.jobStore = f"aws:{self.awsRegion()}:cluster-{uuid4()}"

    @needs_env_var("CI_COMMIT_SHA", "a git commit sha")
    def test_cwl_on_arm(self) -> None:
        # Make a cluster
        self.launchCluster()
        # get the leader so we know the IP address - we don't need to wait since create cluster
        # already ensures the leader is running
        self.cluster = cluster_factory(
            provisioner="aws", zone=self.zone, clusterName=self.clusterName
        )
        self.leader = self.cluster.getLeader()

        commit = os.environ["CI_COMMIT_SHA"]
        self.sshUtil(
            [
                "bash",
                "-c",
                f"mkdir -p {self.cwl_test_dir} && cd {self.cwl_test_dir} && git clone https://github.com/DataBiosphere/toil.git",
            ]
        )

        # We use CI_COMMIT_SHA to retrieve the Toil version needed to run the CWL tests
        self.sshUtil(
            ["bash", "-c", f"cd {self.cwl_test_dir}/toil && git checkout {commit}"]
        )

        # --never-download prevents silent upgrades to pip, wheel and setuptools
        self.sshUtil(
            [
                "bash",
                "-c",
                f"virtualenv --system-site-packages --never-download {self.venvDir}",
            ]
        )
        self.sshUtil(
            [
                "bash",
                "-c",
                f". .{self.venvDir}/bin/activate && cd {self.cwl_test_dir}/toil && make prepare && make develop extras=[all]",
            ]
        )

        # Runs the TestCWLv12 on an ARM instance
        self.sshUtil(
            [
                "bash",
                "-c",
                f". .{self.venvDir}/bin/activate && cd {self.cwl_test_dir}/toil && pytest --log-cli-level DEBUG -r s src/toil/test/cwl/cwlTest.py::TestCWLv12::test_run_conformance",
            ]
        )

        # We know if it succeeds it should save a junit XML for us to read.
        # Bring it back to be an artifact.
        self.rsync_util(
            f":{self.cwl_test_dir}/toil/conformance-1.2.junit.xml",
            str(self._rootpath / "arm-conformance-1.2.junit.xml"),
        )



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pathlib import Path

from toil.job import Job, Promise
from toil.test.src.jobTest import fn1Test


class TestJobEncapsulation:
    """Tests testing the EncapsulationJob class."""

    def testEncapsulation(self, tmp_path: Path) -> None:
        """
        Tests the Job.encapsulation method, which uses the EncapsulationJob
        class.
        """
        # Temporary file
        rootDir = tmp_path / "out"
        rootDir.mkdir()
        outFile = rootDir / "out"
        # Encapsulate a job graph
        a = Job.wrapJobFn(encapsulatedJobFn, "A", outFile, name="a").encapsulate(
            name="a-encap"
        )
        # Now add children/follow to the encapsulated graph
        d = Job.wrapFn(fn1Test, a.rv(), outFile, name="d")
        e = Job.wrapFn(fn1Test, d.rv(), outFile, name="e")
        a.addChild(d)
        a.addFollowOn(e)
        # Create the runner for the workflow.
        options = Job.Runner.getDefaultOptions(tmp_path / "jobstore")
        options.logLevel = "INFO"
        # Run the workflow, the return value being the number of failed jobs
        Job.Runner.startToil(a, options)
        # Check output
        assert outFile.read_text() == "ABCDE"

    def testAddChildEncapsulate(self) -> None:
        """
        Make sure that the encapsulate child does not have two parents
        with unique roots.
        """
        # Temporary file
        a = Job.wrapFn(noOp)
        b = Job.wrapFn(noOp)
        a.addChild(b).encapsulate()
        assert len(a.getRootJobs()) == 1


def noOp() -> None:
    pass


def encapsulatedJobFn(job: Job, string: str, outFile: Path) -> Promise:
    a = job.addChildFn(fn1Test, string, outFile, name="inner-a")
    b = a.addFollowOnFn(fn1Test, a.rv(), outFile, name="inner-b")
    return b.rv()



# Copyright (C) 2025 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains logic for generating Toil usage history reports to send back to Toil
HQ (via Dockstore), and for working out if the user wants to send them.
"""

import asyncio
import collections
import io
import json
import logging
import multiprocessing
import os
import pathlib
import subprocess
import sys
import textwrap
from typing import Any, Literal, Optional, TypeVar, Union

from toil.lib.dockstore import (
    send_metrics,
    get_metrics_url,
    pack_workflow_metrics,
    pack_single_task_metrics,
    pack_workflow_task_set_metrics,
    RunExecution,
    TaskExecutions
)
from toil.lib.history import HistoryManager, WorkflowAttemptSummary, JobAttemptSummary
from toil.lib.misc import unix_seconds_to_local_time
from toil.lib.trs import parse_trs_spec
from toil.lib.io import get_toil_home

logger = logging.getLogger(__name__)

def workflow_execution_id(workflow_attempt: WorkflowAttemptSummary) -> str:
    """
    Get the execution ID for a workflow attempt.

    Result will follow Dockstore's rules.

    Deterministic.
    """

    return workflow_attempt.workflow_id.replace("-", "_") + "_attempt_" + str(workflow_attempt.attempt_number)

def workflow_task_set_execution_id(workflow_attempt: WorkflowAttemptSummary) -> str:
    """
    Get the execution ID for a workflow attempt's collection of tasks.

    Result will follow Dockstore's rules.

    Deterministic.
    """

    return workflow_execution_id(workflow_attempt) + "_tasks"

def job_execution_id(job_attempt: JobAttemptSummary) -> str:
    """
    Get the execution ID for a job attempt.

    Result will follow Dockstore's rules.

    Deterministic.
    """

    return job_attempt.id.replace("-", "_")

def get_parsed_trs_spec(workflow_attempt: WorkflowAttemptSummary) -> tuple[str, str]:
    """
    Get the TRS ID and version of the workflow, or raise an error.

    :returns: The TRS ID and the TRS version of the wrokflow run.
    :raises: ValueError if the workflow does not have a TRS spec or if the spec
        does not contain a version.
    """
    # If it's submittable the TRS spec will be filled in.
    if workflow_attempt.workflow_trs_spec is None:
        raise ValueError("Workflow cannot be submitted without a TRS spec")
    trs_id, trs_version = parse_trs_spec(workflow_attempt.workflow_trs_spec)
    if trs_version is None:
        raise ValueError("Workflow stored in history with TRS ID but without TRS version")
    return trs_id, trs_version

class Submission:
    """
    Class holding a package of information to submit to Dockstore, and the
    information needed to mark the workflows/tasks as submitted if it is accepted.

    Acceptance is at the TRS id/version combination level within the
    Submission, since we need one request per TRS ID/version combination.
    """

    def __init__(self) -> None:
        """
        Create a new empty submission.
        """

        # This collects everything by TRS ID and version, and keeps separate lists for workflows and for task sets.
        # Each thing to be submitted comes along with the reference to what in the database to makr submitted when it goes in.
        self.data: dict[tuple[str, str], tuple[list[tuple[RunExecution, str, int]], list[tuple[TaskExecutions, list[str]]]]] = collections.defaultdict(lambda: ([], []))


    def add_workflow_attempt(self, workflow_attempt: WorkflowAttemptSummary) -> None:
        """
        Add a workflow attempt to the submission.

        May raise an exception if the workflow attempt is not well-formed.
        """
        trs_id, trs_version = get_parsed_trs_spec(workflow_attempt)

        # Figure out what kind of job store was used.
        job_store_type: Optional[str] = None
        try:
            from toil.common import Toil
            job_store_type = Toil.parseLocator(workflow_attempt.workflow_job_store)[0]
            if job_store_type not in Toil.JOB_STORE_TYPES:
                # Make sure we don't send typo'd job store types in.
                job_store_type = None
        except RuntimeError:
            # Could not parse the stored locator.
            pass

        # Pack it up
        workflow_metrics = pack_workflow_metrics(
            workflow_execution_id(workflow_attempt),
            workflow_attempt.start_time,
            workflow_attempt.runtime,
            workflow_attempt.succeeded,
            batch_system=workflow_attempt.batch_system,
            caching=workflow_attempt.caching,
            toil_version=workflow_attempt.toil_version,
            python_version=workflow_attempt.python_version,
            platform_system=workflow_attempt.platform_system,
            platform_machine=workflow_attempt.platform_machine,
            job_store_type=job_store_type
        )

        # Add it to the package
        self.data[(trs_id, trs_version)][0].append((workflow_metrics, workflow_attempt.workflow_id, workflow_attempt.attempt_number))

    def add_job_attempts(self, workflow_attempt: WorkflowAttemptSummary, job_attempts: list[JobAttemptSummary]) -> None:
        """
        Add the job attempts for a workflow attempt to the submission.
        """

        trs_id, trs_version = get_parsed_trs_spec(workflow_attempt)

        # Pack it up
        per_task_metrics = [
            pack_single_task_metrics(
                job_execution_id(job_attempt),
                job_attempt.start_time,
                job_attempt.runtime,
                job_attempt.succeeded,
                job_name=job_attempt.job_name,
                cores=job_attempt.cores,
                cpu_seconds=job_attempt.cpu_seconds,
                memory_bytes=job_attempt.memory_bytes,
                disk_bytes=job_attempt.disk_bytes
            ) for job_attempt in job_attempts
        ]
        task_set_metrics = pack_workflow_task_set_metrics(workflow_task_set_execution_id(workflow_attempt), workflow_attempt.start_time, per_task_metrics)

        # Add it to the package
        self.data[(trs_id, trs_version)][1].append((task_set_metrics, [job_attempt.id for job_attempt in job_attempts]))

    def empty(self) -> bool:
        """
        Return True if there is nothing to actually submit.
        """

        for _, (workflow_metrics, task_metrics) in self.data.items():
            # For each TRS ID and version we might have data for
            if len(workflow_metrics) > 0:
                return False
            if len(task_metrics) > 0:
                # Even if there's no tasks in each, we can report that we ran no tasks
                return False

        return True

    def report(self) -> str:
        """
        Compose a multi-line human-readable string report about what will be submitted.
        """

        stream = io.StringIO()

        for (trs_id, trs_version), (workflow_metrics, task_metrics) in self.data.items():
            stream.write(f"For workflow {trs_id} version {trs_version}:\n")
            if len(workflow_metrics) > 0:
                stream.write("  Workflow executions:\n")
                for workflow_metric, _, _ in workflow_metrics:
                    stream.write(textwrap.indent(json.dumps(workflow_metric, indent=2), '    '))
            if len(task_metrics) > 0:
                stream.write("  Task execution sets:\n")
                for tasks_metric, _ in task_metrics:
                    stream.write(textwrap.indent(json.dumps(tasks_metric, indent=2), '    '))

        return stream.getvalue()

    def submit(self) -> bool:
        """
        Send in workflow and task information to Dockstore.

        Assumes the user has approved this already.

        If submission succeeds for a TRS id/version combination, records this in the history database.

        Handles errors internally. Will not raise if the submission doesn't go through.

        :returns: False if there was any error, True if submission was accepted and recorded locally.
        """

        all_submitted_and_marked = True

        for (trs_id, trs_version), (workflow_metrics, task_metrics) in self.data.items():
            submitted = False
            try:
                send_metrics(trs_id, trs_version, [item[0] for item in workflow_metrics], [item[0] for item in task_metrics])
                submitted = True
            except:
                logger.exception("Could not submit to Dockstore")
                all_submitted_and_marked = False
                # Ignore failed submissions and keep working on other ones.

            if submitted:
                for _, workflow_id, attempt_number in workflow_metrics:
                    # For each workflow attempt of this TRS ID/version that we successfully submitted, mark it
                    try:
                        HistoryManager.mark_workflow_attempt_submitted(workflow_id, attempt_number)
                    except:
                        logger.exception("Could not mark workflow attempt submitted")
                        all_submitted_and_marked = False
                        # If we can't mark one, keep marking others
                for _, job_attempt_ids in task_metrics:
                    # For each colleciton of task attempts of this TRS ID/version that we successfully submitted, mark it
                    try:
                        HistoryManager.mark_job_attempts_submitted(job_attempt_ids)
                    except:
                        logger.exception("Could not mark job attempts submitted")
                        all_submitted_and_marked = False
                        # If we can't mark one set, keep marking others

        return all_submitted_and_marked


def create_history_submission(batch_size: Optional[int] = None, desired_tasks: Optional[int] = None) -> Submission:
    """
    Make a package of data about recent workflow runs to send in.

    Returns a Submission object that remembers how to update the history
    database to record when it is successfully submitted.

    :param batch_size: Number of workflows to try and submit in one request.
    :param desired_tasks: Number of tasks to try and put into a task submission
        batch. Use 0 to not submit any task information.
    """

    # By default, include the things we are set to track history for.
    if batch_size is None:
        batch_size = 10 if HistoryManager.enabled() else 0
    if desired_tasks is None:
        desired_tasks = 50 if HistoryManager.enabled_job() else 0

    # Collect together some workflows and some lists of tasks into a submission.
    submission = Submission()

    # We don't want to submit many large workflows' tasks in one request. So we
    # count how many tasks we've colected so far.
    total_tasks = 0

    for workflow_attempt in HistoryManager.get_submittable_workflow_attempts(limit=batch_size):
        try:
            submission.add_workflow_attempt(workflow_attempt)
        except:
            logger.exception("Could not compose metrics report for workflow execution")
            # Ignore failed submissions and keep working on other ones.
            # TODO: What happens when they accumulate and fill the whole batch from the database?

        # TODO: Dockstore limits request size to 60k. See
        # <https://github.com/dockstore/compose_setup/blob/a14cd1d390f4ae1e14a8ba6e36ec06ce5fe2604e/templates/default.nginx_http.shared.conf.template#L15>.
        # Apply that limit to the submission and don't add more attempts if we
        # would go over it.

    # TODO: Once we have a way to deal with single runs possibly having more
    # than 60k worth of tasks, and with Dockstore turning collections of tasks
    # into extra workflow runs (see
    # <https://ucsc-cgl.atlassian.net/browse/SEAB-6919>), set the default task
    # limit to submit to be nonzero.

    for workflow_attempt in HistoryManager.get_workflow_attempts_with_submittable_job_attempts(limit=batch_size):
        job_attempts = HistoryManager.get_unsubmitted_job_attempts(workflow_attempt.workflow_id, workflow_attempt.attempt_number)

        if desired_tasks == 0 or total_tasks + len(job_attempts) > desired_tasks:
            # Don't add any more task sets to the submission
            break

        try:
            submission.add_job_attempts(workflow_attempt, job_attempts)

            total_tasks += len(job_attempts)
        except:
            logger.exception("Could not compose metrics report for workflow task set")
            # Ignore failed submissions and keep working on other ones.
            # TODO: What happens when they accumulate and fill the whole batch from the database?

    return submission

def create_current_submission(workflow_id: str, attempt_number: int) -> Submission:
    """
    Make a package of data about the current workflow attempt to send in.

    Useful if the user wants to submit workflow metrics just this time.
    """

    submission = Submission()
    try:
        workflow_attempt = HistoryManager.get_workflow_attempt(workflow_id, attempt_number)
        if workflow_attempt is not None and HistoryManager.enabled():
            if not workflow_attempt.submitted_to_dockstore:
                submission.add_workflow_attempt(workflow_attempt)
            if HistoryManager.enabled_job():
                try:
                    job_attempts = HistoryManager.get_unsubmitted_job_attempts(workflow_attempt.workflow_id, workflow_attempt.attempt_number)
                    submission.add_job_attempts(workflow_attempt, job_attempts)
                except:
                    logger.exception("Could not compose metrics report for workflow task set")
                    # Keep going with just the workflow.
    except:
        logger.exception("Could not compose metrics report for workflow execution")
        # Keep going with an empty submission.

    return submission

# We have dialog functions that MyPy knows can return strings from a possibly restricted set
KeyType = TypeVar('KeyType', bound=str)

def dialog_tkinter(title: str, text: str, options: dict[KeyType, str], timeout: float) -> Optional[KeyType]:
    """
    Display a dialog with tkinter.

    Dialog will have the given title, text, and options.

    Dialog will be displayed by a separate Python process to avoid a Mac dock
    icon sticking around as long as Toil runs.

    :param options: Dict from machine-readable option key to button text.
    :returns: the key of the selected option, or None if the user declined to
        select an option.
    :raises: an exception if the dialog cannot be displayed.
    """

    # Multiprocessing queues aren't actually generic, but MyPy requires them to be.
    result_queue: "multiprocessing.Queue[Union[Exception, Optional[KeyType]]]" = multiprocessing.Queue()
    process = multiprocessing.Process(target=display_dialog_tkinter, args=(title, text, options, timeout, result_queue))
    process.start()
    result = result_queue.get()
    process.join()

    if isinstance(result, Exception):
        raise result
    else:
        return result



def display_dialog_tkinter(title: str, text: str, options: dict[KeyType, str], timeout: float, result_queue: "multiprocessing.Queue[Union[Exception, Optional[KeyType]]]") -> None:
    """
    Display a dialog with tkinter in the current process.

    Dialog will have the given title, text, and options.

    :param options: Dict from machine-readable option key to button text.

    Sends back either the key of the chosen option, or an exception, via the
    result queue.
    """

    try:

        FRAME_PADDING = 10
        BUTTON_FRAME_PADDING = 5
        DEFAULT_LABEL_WRAP = 400

        import tkinter
        from tkinter import ttk

        # Get a root window
        root = tkinter.Tk()
        # Set the title
        root.title(title)

        def close_root() -> None:
            """
            Function to close the dialog window.
            """
            # Hide the window
            root.withdraw()
            # Now we want to exit the main loop. But if we do it right away the window hide never happens and the window stays open but not responding.
            # So we schedule the root to go away.
            root.after(100, root.destroy)


        # Make a frame
        frame = ttk.Frame(root, padding=FRAME_PADDING)
        # Put it on a grid in the parent
        frame.grid(row=0, column=0, sticky="nsew")

        # Lay out a label in the frame's grid
        label = ttk.Label(frame, text=text, wraplength=DEFAULT_LABEL_WRAP)
        label.grid(column=0, row=0, sticky="nsew")

        # Add a button frame below it so the buttons can be in columns narrower than the label
        button_frame = ttk.Frame(frame, padding=BUTTON_FRAME_PADDING)
        # Put it on a grid in the parent
        button_frame.grid(row=1, column=0, sticky="sw")

        # Use a list as a mutable slot
        result: list[KeyType] = []
        button_column = 0
        for k, v in options.items():
            # A function in here will capture k by reference. Sneak the current k
            # in through a default argument, which is evaluated at definition time.
            def setter(set_to: KeyType = k) -> None:
                # Record the choice
                result.append(set_to)
                # Close the window.
                close_root()
            ttk.Button(button_frame, text=v, command=setter).grid(column=button_column, row=0)
            button_column += 1

        # Buttons do not grow as the window resizes.

        # Say row 0 of the container frame gets all its extra height
        frame.rowconfigure(0, weight=1)
        # Say column 0 of the container frame gets all its extra width
        frame.columnconfigure(0, weight=1)

        # Say row 0 of the root gets all its extra height
        root.rowconfigure(0, weight=1)
        # Say column 0 of the root gets all its extra width
        root.columnconfigure(0, weight=1)

        # Make sure the window can't get too small for the content
        def force_fit() -> None:
            """
            Make sure the root window is no smaller than its contents require.
            """
            # Give it a minimum size in inches, and also no smaller than the
            # contents require. We can't use the required width exactly because
            # that would never let us shrink and rewrap the label, so we use the
            # button frame instead and add padding ourselves.
            min_width = max(int(root.winfo_fpixels("2i")), button_frame.winfo_reqwidth() + BUTTON_FRAME_PADDING + FRAME_PADDING)
            min_height = max(int(root.winfo_fpixels("1i")), root.winfo_reqheight())
            root.minsize(min_width, min_height)

            # TODO: If the window would be too tall for the screen, make it wider for the user?

        # Rewrap the text as the window size changes
        setattr(label, "last_width", 100)
        # MyPy demands a generic argument here but Python won't be able to handle
        # it until
        # <https://github.com/python/cpython/commit/42a818912bdb367c4ec2b7d58c18db35f55ebe3b>
        # ships.
        def resize_label(event: "tkinter.Event[ttk.Label]") -> None:
            """
            If the label's width has changed, rewrap the text.

            See <https://stackoverflow.com/a/71599924>
            """
            current_width = label.winfo_width()
            if getattr(label, "last_width") != current_width:
                setattr(label, "last_width", current_width)
                label.config(wraplength=current_width)
                # Update required height calculation based on new width
                label.update_idletasks()
                # Impose minimum height
                force_fit()
        label.bind('<Configure>', resize_label)

        # Do root sizing
        force_fit()

        if timeout:
            # If we run out of time, hide the window and move on without a choice.
            root.after(int(timeout * 1000), close_root)

        # To make the dialog pop up over the terminal instead of behind it, we
        # lift it and temporarily make it topmost. We don't keep it topmost
        # because we want to let the user switch away from it.
        root.attributes('-topmost', True)
        root.after(10, lambda: root.attributes('-topmost', False))
        root.lift()

        # Run the window's main loop
        root.mainloop()

        if len(result) == 0:
            # User didn't click any buttons. Probably they don't want to deal with
            # this right now.
            result_queue.put(None)
        else:
            result_queue.put(result[0])
    except Exception as e:
        result_queue.put(e)

def dialog_tui(title: str, text: str, options: dict[KeyType, str], timeout: float) -> Optional[KeyType]:
    """
    Display a dialog in the terminal.

    Dialog will have the given title, text, and options.

    :param options: Dict from machine-readable option key to button text.
    :returns: the key of the selected option, or None if the user declined to
        select an option.
    :raises: an exception if the dialog cannot be displayed.
    """

    # See
    # <https://python-prompt-toolkit.readthedocs.io/en/master/pages/dialogs.html#button-dialog>
    # for the dialog and
    # <https://python-prompt-toolkit.readthedocs.io/en/master/pages/advanced_topics/input_hooks.html>
    # for how we run concurrently with it.

    from prompt_toolkit.shortcuts import button_dialog
    from prompt_toolkit.eventloop.inputhook import set_eventloop_with_inputhook

    # We take button options in the reverse order from how prompt_toolkit does it.
    # TODO: This is not scrollable! What if there's more than 1 screen of text?
    # TODO: Use come kind of ScrollablePane like <https://stackoverflow.com/q/68369073>
    application = button_dialog(
        title=title,
        text=text,
        buttons=[(v, k) for k, v in options.items()],
    )

    # Make the coroutine to run the dialog
    application_coroutine = application.run_async()

    async def exit_application_after_timeout() -> None:
        """
        Wait for the timeout to elapse and then close the application.
        """
        await asyncio.sleep(timeout)
        # End the application and make it return None
        application.exit()

    # Make the coroutine to stop the dialog
    timeout_coroutine = exit_application_after_timeout()

    # To avoid messing around with a global event loop and instead just race
    # some coroutines, we make our own event loop.
    loop = asyncio.new_event_loop()

    # Attach the coroutines to it so it tries to advance them
    application_task = loop.create_task(application_coroutine)
    timeout_task = loop.create_task(timeout_coroutine)

    # This task finishes when the first task in the set finishes.
    race_task = asyncio.wait({application_task, timeout_task}, return_when=asyncio.FIRST_COMPLETED)
    # Run it until then
    loop.run_until_complete(race_task)
    # Maybe the application needs to finish its exit?
    loop.run_until_complete(application_task)

    # We no longer need the stopping task/coroutine
    timeout_task.cancel()
    # We don't need to await it because we cancel it.
    # It is OK to cancel it if it is finished already.

    return application_task.result()

# Define the dialog form in the abstract
Decision = Union[Literal["all"], Literal["current"], Literal["no"], Literal["never"]]

DIALOG_TITLE = "Publish Workflow Metrics on Dockstore.org?"

# We need to fromat the default config path in to the message, but we can't get
# it until we can close the circular import loop. So it's a placeholder here.
# We also leave this un-wrapped to let the dialog wrap it if needed.
DIALOG_TEXT = """
Would you like to publish execution metrics on Dockstore.org?

This includes information like a unique ID for the workflow execution and each job execution, the Tool Registry Service (TRS) ID of the workflow, the names of its jobs, when and for how long they run, how much CPU, memory, and disk they are allocated or use, whether they succeed or fail, the versions of Toil and Python used, the operating system platform and processor type, and which Toil or Toil plugin features are used.

Dockstore.org uses this information to prepare reports about how well workflows run in different environments, and what resources they need, in order to help users plan their workflow runs. The Toil developers also consult this information to see which Toil features are the most popular and how popular Toil is overall.

Note that publishing is PERMANENT! You WILL NOT be able to recall or un-publish any published metrics!

(You can change your choice by editing the "publishWorkflowMetrics" setting in the "{}" file. You can override it once with the "--publishWorkflowMetrics" command line option.)

All: Publish for this run and all future AND PAST runs of ALL workflows.
Yes: Publish for just this run, and ask again next time.
No: Do not publish anything now, but ask again next time.
Never: Do not publish anything and stop asking.
""".strip()

DIALOG_OPTIONS: dict[Decision, str] = {
    "all": "All",
    "current": "Yes",
    "no": "No",
    "never": "Never"
}

# Make sure the option texts are short enough; prompt_toolkit can only handle 10 characters.
for k, v in DIALOG_OPTIONS.items():
    assert len(v) <= 10, f"Label for \"{k}\" dialog option is too long to work on all backends!"

# Make sure the options all have unique labels
assert len(set(DIALOG_OPTIONS.values())) == len(DIALOG_OPTIONS), "Labels for dialog options are not unique!"

# How many seconds should we show the dialog for before assuming the user means "no"?
DIALOG_TIMEOUT = 120

def ask_user_about_publishing_metrics() -> Union[Literal["all"], Literal["current"], Literal["no"]]:
    """
    Ask the user to set standing workflow submission consent.

    If the user makes a persistent decision (always or never), save it to the default Toil config file.

    :returns: The user's decision about when to publish metrics.
    """

    from toil.common import update_config, get_default_config_path

    # Find the default config path to talk about or update
    default_config_path = get_default_config_path()

    # Actual chatting with the user will fill this in if possible
    decision: Optional[Decision] = None

    if sys.stdin.isatty() and sys.stderr.isatty():
        # IO is not redirected (except maybe JSON to a file). We might be able to raise the user.

        # TODO: Get a lock

        for strategy in [dialog_tkinter, dialog_tui]:
            # TODO: Add a graphical strategy that, unlike tkinter (which needs
            # homebrew python-tk), can work out of the box on Mac! AppleScript
            # and builtin NSDialog can only use 3 buttons. AppleScript choice
            # dialogs can't hold enough text. We don't really want to bring in
            # e.g. QT for this.
            try:
                strategy_decision = strategy(DIALOG_TITLE, DIALOG_TEXT.format(default_config_path), DIALOG_OPTIONS, DIALOG_TIMEOUT)
                if strategy_decision is None:
                    # User declined to choose. Treat that as choosing "no".
                    logger.warning("User did not make a selection. Assuming \"no\".")
                    strategy_decision = "no"
                decision = strategy_decision
                break
            except:
                logger.exception("Could not use %s", strategy)

        if decision is None:
            # If we think we should be able to reach the user, but we can't, fail and make them tell us via option
            # TODO: Provide a command (or a general toil config editing command) to manage this setting
            logger.critical("Decide whether to publish workflow metrics and pass --publishWorkflowMetrics=[all|current|no]")
            sys.exit(1)

    if decision is None:
        # We can't reach the user
        decision = "no"

    result = decision if decision != "never" else "no"
    if decision in ("all", "never"):
        # These are persistent and should save to the config
        update_config(default_config_path, "publishWorkflowMetrics", result)

    return result














#!/usr/bin/env python3
"""
A script to remove all trailing whitespace from the Toil codebase's text files
(mainly targeting .py files, but also things like .sh and .txt files).

Run with:
    remove_trailing_whitespace.py

Note: This script preserves file permissions (and thus, executability) and that should be maintained.
"""
import os
import sys

pkg_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))  # noqa
sys.path.insert(0, pkg_root)  # noqa


# changing either of these variables will change the targets of this script
TOIL_DIRS_TO_PROCESS = {'attic', 'dashboard', 'docker', 'docs', 'src'}
EXTENSIONS_TO_PROCESS = {'.csv',
                         '.cwl',
                         '.input',
                         '.json',
                         '.md',
                         '.mtail',
                         '.py',
                         '.rst',
                         '.sh',
                         '.test-pr',
                         '.tsv',
                         '.txt',
                         '.wdl',
                         '.yaml',
                         '.yml',
                         'Dockerfile',
                         'Makefile'}


def strip_trailing_whitespace_from_file(filename: str) -> None:
    """Strips trailing whitespace from a file, in-place."""
    pristine_lines = []
    with open(filename, 'r') as r:
        for line in r:
            pristine_lines.append(line.rstrip())

    with open(filename, 'w') as w:
        for pristine_line in pristine_lines:
            w.write(pristine_line + '\n')


def strip_trailing_whitespace_from_all_files_in_dir(dirname: str) -> None:
    """
    Strips trailing whitespace from all files in a directory, recursively.

    Only strips files ending in one of the white-listed extensions in EXTENSIONS_TO_PROCESS.
    Note: This includes things like "Dockerfile" that end in "Dockerfile".
    """
    for dirpath, dirnames, filenames in os.walk(dirname):
        for f in filenames:
            for ext in EXTENSIONS_TO_PROCESS:
                if f.endswith(ext):
                    strip_trailing_whitespace_from_file(os.path.abspath(os.path.join(dirpath, f)))
                    break
        for d in dirnames:
            strip_trailing_whitespace_from_all_files_in_dir(os.path.abspath(os.path.join(dirpath, d)))


def main():
    for directory in TOIL_DIRS_TO_PROCESS:
        strip_trailing_whitespace_from_all_files_in_dir(os.path.join(pkg_root, directory))


if __name__ == '__main__':
    main()



import os

from toil.common import Toil
from toil.job import Job
from toil.lib.io import mkdtemp


def binaryStringFn(job, depth, message=""):
    if depth > 0:
        job.addChildJobFn(binaryStringFn, depth - 1, message + "0")
        job.addChildJobFn(binaryStringFn, depth - 1, message + "1")
    else:
        job.log(f"Binary string: {message}")


if __name__ == "__main__":
    jobstore: str = mkdtemp("tutorial_dynamic")
    os.rmdir(jobstore)
    options = Job.Runner.getDefaultOptions(jobstore)
    options.logLevel = "INFO"
    options.clean = "always"

    with Toil(options) as toil:
        toil.start(Job.wrapJobFn(binaryStringFn, depth=5))



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
import os
import shutil
import uuid
from collections import Counter
from collections.abc import Generator
from contextlib import contextmanager
from typing import Any, Callable, Optional, TextIO, Union, overload

from flask import send_from_directory
from werkzeug.utils import redirect
from werkzeug.wrappers.response import Response

import toil.server.wes.amazon_wes_utils as amazon_wes_utils
from toil.bus import JobStatus, replay_message_bus
from toil.lib.io import AtomicFileCreate
from toil.lib.threading import global_mutex
from toil.server.utils import WorkflowStateMachine, connect_to_workflow_state_store
from toil.server.wes.abstract_backend import (
    OperationForbidden,
    TaskLog,
    VersionNotImplementedException,
    WESBackend,
    WorkflowConflictException,
    WorkflowExecutionException,
    WorkflowNotFoundException,
    handle_errors,
)
from toil.server.wes.tasks import MultiprocessingTaskRunner, TaskRunner
from toil.version import baseVersion

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class ToilWorkflow:
    def __init__(self, base_work_dir: str, state_store_url: str, run_id: str):
        """
        Class to represent a Toil workflow. This class is responsible for
        launching workflow runs via Celery and retrieving data generated from
        them.

        :param base_work_dir: The directory where workflows keep their
                                 output/scratch directories under their run
                                 IDs. Includes the MessageBus which requires
                                 a reliable location.

        :param state_store_url: URL or file path at which we communicate with
                                running workflows.

        :param run_id: A unique per-run string.  Used to name the folder that contains
                       all of the files containing this particular workflow
                       instance's information.
        """
        self.base_scratch_dir = base_work_dir
        self.state_store_url = state_store_url
        self.run_id = run_id

        self.scratch_dir = os.path.join(self.base_scratch_dir, self.run_id)
        self.exec_dir = os.path.join(self.scratch_dir, "execution")

        # TODO: share a base class with ToilWorkflowRunner for some of this stuff?

        # Connect to the workflow state store
        self.store = connect_to_workflow_state_store(state_store_url, self.run_id)
        # And use a state machine over that to look at workflow state
        self.state_machine = WorkflowStateMachine(self.store)

    @overload
    def fetch_state(self, key: str, default: str) -> str: ...
    @overload
    def fetch_state(self, key: str, default: None = None) -> Optional[str]: ...

    def fetch_state(self, key: str, default: Optional[str] = None) -> Optional[str]:
        """
        Return the contents of the given key in the workflow's state
        store. If the key does not exist, the default value is returned.
        """
        value = self.store.get(key)
        if value is None:
            return default
        return value

    @contextmanager
    def fetch_scratch(self, filename: str) -> Generator[Optional[TextIO], None, None]:
        """
        Get a context manager for either a stream for the given file from the
        workflow's scratch directory, or None if it isn't there.
        """
        if os.path.exists(os.path.join(self.scratch_dir, filename)):
            with open(os.path.join(self.scratch_dir, filename)) as f:
                yield f
        else:
            yield None

    def exists(self) -> bool:
        """Return True if the workflow run exists."""
        return self.get_state() != "UNKNOWN"

    def get_state(self) -> str:
        """Return the state of the current run."""
        return self.state_machine.get_current_state()

    def check_on_run(self, task_runner: type[TaskRunner]) -> None:
        """
        Check to make sure nothing has gone wrong in the task runner for this
        workflow. If something has, log, and fail the workflow with an error.
        """
        if not task_runner.is_ok(self.run_id) and self.get_state() not in [
            "SYSTEM_ERROR",
            "EXECUTOR_ERROR",
            "COMPLETE",
            "CANCELED",
        ]:
            logger.error(
                "Failing run %s because the task to run its leader crashed", self.run_id
            )
            self.state_machine.send_system_error()

    def set_up_run(self) -> None:
        """Set up necessary directories for the run."""
        # Go to queued state
        self.state_machine.send_enqueue()

        # Make sure scratch and exec directories exist
        os.makedirs(self.exec_dir, exist_ok=True)

    def clean_up(self) -> None:
        """Clean directory and files related to the run."""
        shutil.rmtree(self.scratch_dir)
        # Don't remove state; state needs to persist forever.

    def queue_run(
        self, task_runner: type[TaskRunner], request: dict[str, Any], options: list[str]
    ) -> None:
        """This workflow should be ready to run. Hand this to the task system."""
        with open(os.path.join(self.scratch_dir, "request.json"), "w") as f:
            # Save the request to disk for get_run_log()
            json.dump(request, f)

        try:
            # Run the task. Set the task ID the same as our run ID
            task_runner.run(
                args=(
                    self.base_scratch_dir,
                    self.state_store_url,
                    self.run_id,
                    request,
                    options,
                ),
                task_id=self.run_id,
            )
        except Exception:
            # Celery or the broker might be down
            self.state_machine.send_system_error()
            raise WorkflowExecutionException(f"Failed to run: internal server error.")

    def get_output_files(self) -> Any:
        """
        Return a collection of output files that this workflow generated.
        """
        with self.fetch_scratch("outputs.json") as f:
            if f is None:
                # No file is there
                return {}
            else:
                # Stream in the file
                return json.load(f)

    def _get_scratch_file_path(self, path: str) -> Optional[str]:
        """
        Return the given relative path from self.scratch_dir, if it is a file,
        and None otherwise.
        """
        if not os.path.isfile(os.path.join(self.scratch_dir, path)):
            return None
        return path

    def get_stdout_path(self) -> Optional[str]:
        """
        Return the path to the standard output log, relative to the run's
        scratch_dir, or None if it doesn't exist.
        """
        return self._get_scratch_file_path("stdout")

    def get_stderr_path(self) -> Optional[str]:
        """
        Return the path to the standard output log, relative to the run's
        scratch_dir, or None if it doesn't exist.
        """
        return self._get_scratch_file_path("stderr")

    def get_messages_path(self) -> Optional[str]:
        """
        Return the path to the bus message log, relative to the run's
        scratch_dir, or None if it doesn't exist.
        """
        return self._get_scratch_file_path("bus_messages")

    def get_task_logs(
        self,
        filter_function: Optional[
            Callable[[TaskLog, JobStatus], Optional[TaskLog]]
        ] = None,
    ) -> list[dict[str, Union[str, int, None]]]:
        """
        Return all the task log objects for the individual tasks in the workflow.

        Task names will be the job_type values from issued/completed/failed
        messages, with annotations from JobAnnotationMessage messages if
        available.

        :param filter_function: If set, will be called with each task log and
               its job annotations. Returns a modified copy of the task log to
               actually report, or None if the task log should be omitted.
        """

        # First, find where we kept the message log from the workflow that we
        # can actually decode.
        path = self.get_messages_path()

        if path is None:
            # No task logs available
            return []
        else:
            # Replay all the messages and work out what they mean for jobs.
            abs_path = os.path.join(self.scratch_dir, path)
            job_statuses = replay_message_bus(abs_path)
            # Compose log objects from recovered job info.
            logs: list[TaskLog] = []
            for job_status in job_statuses.values():
                task: Optional[TaskLog] = {
                    "name": job_status.name,
                    "exit_code": job_status.exit_code,
                }
                if filter_function is not None:
                    # Convince MyPy the task is set
                    assert task is not None
                    # Give the filter function hook a chance to modify or omit the task
                    task = filter_function(task, job_status)
                if task is not None:
                    logs.append(task)
            logger.info("Recovered task logs: %s", logs)
            return logs
            # TODO: times, log files, AWS Batch IDs if any, names from the workflow instead of IDs, commands


class ToilBackend(WESBackend):
    """
    WES backend implemented for Toil to run CWL, WDL, or Toil workflows. This
    class is responsible for validating and executing submitted workflows.
    """

    def __init__(
        self,
        work_dir: str,
        state_store: Optional[str],
        options: list[str],
        dest_bucket_base: Optional[str],
        bypass_celery: bool = False,
        wes_dialect: str = "standard",
    ) -> None:
        """
        Make a new ToilBackend for serving WES.

        :param work_dir: Directory to download and run workflows in.

        :param state_store: Path or URL to store workflow state at.

        :param options: Command-line options to pass along to workflows. Must
               use = syntax to set values instead of ordering.

        :param dest_bucket_base: If specified, direct CWL workflows to use
               paths under the given URL for storing output files.

        :param bypass_celery: Can be set to True to bypass Celery and the
               message broker and invoke workflow-running tasks without them.

        :param wes_dialect: Can be set to "agc" to emit WES responses that are
               compatible with Amazon Genomics CLI's restricted subset of
               acceptable WES responses. Requests are always accepted if
               acceptable in any dialect.
        """
        for opt in options:
            if not opt.startswith("-"):
                # We don't allow a value to be set across multiple arguments
                # that would need to remain in the same order.
                raise ValueError(f"Option {opt} does not begin with -")
        super().__init__(options)

        # How should we generate run IDs? We apply a prefix so that we can tell
        # what things in our work directory suggest that runs exist and what
        # things don't.
        self.run_id_prefix = "run-"

        # Use this to run Celery tasks so we can swap it out for testing.
        self.task_runner = (
            TaskRunner if not bypass_celery else MultiprocessingTaskRunner
        )
        logger.info("Using task runner: %s", self.task_runner)

        # Record if we need to limit our WES responses for a particular
        # non-compliant consumer
        self.wes_dialect = wes_dialect
        logger.info("Using WES dialect: %s", self.wes_dialect)

        self.dest_bucket_base = dest_bucket_base
        logger.info("Using destination bucket base URL: %s", self.dest_bucket_base)
        self.work_dir = os.path.abspath(work_dir)
        logger.info("Using work dir: %s", self.work_dir)
        os.makedirs(self.work_dir, exist_ok=True)

        # Where should we talk to the tasks about workflow state?

        if state_store is None:
            # Store workflow metadata under the work_dir.
            self.state_store_url = os.path.join(self.work_dir, "state_store")
        else:
            # Use the provided value
            self.state_store_url = state_store
        logger.info("Using state store: %s", self.state_store_url)

        # Determine a server identity, so we can guess if a workflow in the
        # possibly-persistent state store is QUEUED, INITIALIZING, or RUNNING
        # on a Celery that no longer exists. Ideally we would ask Celery what
        # Celery cluster it is, or we would reconcile with the tasks that exist
        # in the Celery cluster, but Celery doesn't seem to have a cluster
        # identity and doesn't let you poll for task existence:
        # <https://stackoverflow.com/questions/9824172>

        # Grab an ID for the current kernel boot, if we happen to be Linux.
        # TODO: Deal with multiple servers in front of the same state store and
        # file system but on different machines?
        boot_id = None
        boot_id_file = "/proc/sys/kernel/random/boot_id"
        if os.path.exists(boot_id_file):
            try:
                with open(boot_id_file) as f:
                    boot_id = f.readline().strip()
            except OSError:
                pass
        # Assign an ID to the work directory storage.
        work_dir_id = None
        work_dir_id_file = os.path.join(self.work_dir, "id.txt")
        if os.path.exists(work_dir_id_file):
            # An ID is assigned already
            with open(work_dir_id_file) as f:
                work_dir_id = uuid.UUID(f.readline().strip())
        else:
            # We need to try and assign an ID.
            with global_mutex(self.work_dir, "id-assignment"):
                # We need to synchronize with other processes starting up to
                # make sure we agree on an ID.
                if os.path.exists(work_dir_id_file):
                    # An ID is assigned already
                    with open(work_dir_id_file) as f:
                        work_dir_id = uuid.UUID(f.readline().strip())
                else:
                    work_dir_id = uuid.uuid4()
                    with AtomicFileCreate(work_dir_id_file) as temp_file:
                        # Still need to be atomic here or people not locking
                        # will see an incomplete file.
                        with open(temp_file, "w") as f:
                            f.write(str(work_dir_id))
        # Now combine into one ID
        if boot_id is not None:
            self.server_id = str(uuid.uuid5(work_dir_id, boot_id))
        else:
            self.server_id = str(work_dir_id)
        logger.info("Using server ID: %s", self.server_id)

        self.supported_versions = {
            "py": ["3.7", "3.8", "3.9"],
            "cwl": ["v1.0", "v1.1", "v1.2"],
            "wdl": ["draft-2", "1.0"],
        }

    def _get_run(
        self, run_id: str, should_exists: Optional[bool] = None
    ) -> ToilWorkflow:
        """
        Helper method to instantiate a ToilWorkflow object.

        :param run_id: The run ID.
        :param should_exists: If set, ensures that the workflow run exists (or
                              does not exist) according to the value.
        """
        run = ToilWorkflow(self.work_dir, self.state_store_url, run_id)

        if should_exists and not run.exists():
            raise WorkflowNotFoundException
        if should_exists is False and run.exists():
            raise WorkflowConflictException(run_id)

        # Do a little fixup of orphaned/Martian workflows that were running
        # before the server bounced and can't be running now.
        # Sadly we can't just ask Celery if it has heard of them.
        # TODO: Implement multiple servers working together.
        owning_server = run.fetch_state("server_id")
        apparent_state = run.get_state()
        if (
            apparent_state
            not in ("UNKNOWN", "COMPLETE", "EXECUTOR_ERROR", "SYSTEM_ERROR", "CANCELED")
            and owning_server != self.server_id
        ):

            # This workflow is in a state that suggests it is doing something
            # but it appears to belong to a previous incarnation of the server,
            # and so its Celery is probably gone. Put it into system error
            # state if possible.
            logger.warning(
                "Run %s in state %s appears to belong to server %s and not us, server %s. "
                "Its server is probably gone. Failing the workflow!",
                run_id,
                apparent_state,
                owning_server,
                self.server_id,
            )
            run.state_machine.send_system_error()

        # Poll to make sure the run is not broken
        run.check_on_run(self.task_runner)
        return run

    def get_runs(self) -> Generator[tuple[str, str], None, None]:
        """A generator of a list of run ids and their state."""
        if not os.path.exists(self.work_dir):
            return

        for run_id in os.listdir(self.work_dir):
            if not run_id.startswith(self.run_id_prefix):
                continue
            run = self._get_run(run_id)
            if run.exists():
                yield run_id, run.get_state()

    def get_state(self, run_id: str) -> str:
        """
        Return the state of the workflow run with the given run ID. May raise
        an error if the workflow does not exist.
        """
        return self._get_run(run_id, should_exists=True).get_state()

    @handle_errors
    def get_service_info(self) -> dict[str, Any]:
        """Get information about the Workflow Execution Service."""

        state_counts = Counter(state for _, state in self.get_runs())

        engine_parameters = []
        for option in self.options:
            if "=" not in option:  # flags like "--logDebug"
                k, v = option, None
            else:
                k, v = option.split("=", 1)
            engine_parameters.append((k, v))

        return {
            "version": baseVersion,
            "workflow_type_versions": {
                k: {"workflow_type_version": v}
                for k, v in self.supported_versions.items()
            },
            "supported_wes_versions": ["1.0.0"],
            "supported_filesystem_protocols": ["file", "http", "https"],
            "workflow_engine_versions": {"toil": baseVersion},
            # TODO: How can we report --destBucket here, since we pass it only
            # for CWL workflows?
            "default_workflow_engine_parameters": [
                {"name": key, "default_value": value}
                for key, value in engine_parameters
            ],
            "system_state_counts": state_counts,
            "tags": {"wes_dialect": self.wes_dialect},
        }

    @handle_errors
    def list_runs(
        self, page_size: Optional[int] = None, page_token: Optional[str] = None
    ) -> dict[str, Any]:
        """List the workflow runs."""
        # TODO: implement pagination
        return {
            "workflows": [
                {"run_id": run_id, "state": state} for run_id, state in self.get_runs()
            ],
            "next_page_token": "",
        }

    @handle_errors
    def run_workflow(self) -> dict[str, str]:
        """Run a workflow."""
        run_id = self.run_id_prefix + uuid.uuid4().hex
        run = self._get_run(run_id, should_exists=False)

        # set up necessary directories for the run. We need to do this because
        # we need to save attachments before we can get ahold of the request.
        run.set_up_run()

        # stage the uploaded files to the execution directory, so that we can run the workflow file directly
        temp_dir = run.exec_dir
        try:
            _, request = self.collect_attachments(run_id, temp_dir=temp_dir)
        except ValueError:
            run.clean_up()
            raise

        logger.info(
            "Received workflow run request %s with parameters: %s",
            run_id,
            list(request.keys()),
        )

        wf_type = request["workflow_type"].lower().strip()
        version = request["workflow_type_version"]

        # validate workflow request
        supported_versions = self.supported_versions.get(wf_type, None)
        if not supported_versions:
            run.clean_up()
            raise VersionNotImplementedException(wf_type)
        if version not in supported_versions:
            run.clean_up()
            raise VersionNotImplementedException(wf_type, version, supported_versions)

        # Now we are actually going to try and do the run.

        # Claim ownership of the run
        run.store.set("server_id", self.server_id)

        # Generate workflow options
        workflow_options = list(self.options)
        if wf_type == "cwl" and self.dest_bucket_base:
            # Output to a directory under out base destination bucket URL.
            workflow_options.append(
                "--destBucket=" + os.path.join(self.dest_bucket_base, run_id)
            )
        # Tell it to dump its messages to a file.
        # TODO: automatically sync file names with accessors somehow.
        workflow_options.append(
            "--writeMessages=" + os.path.join(run.scratch_dir, "bus_messages")
        )

        logger.info(
            f"Putting workflow {run_id} into the queue. Waiting to be picked up..."
        )
        run.queue_run(self.task_runner, request, options=workflow_options)

        return {"run_id": run_id}

    @handle_errors
    def get_run_log(self, run_id: str) -> dict[str, Any]:
        """Get detailed info about a workflow run."""
        run = self._get_run(run_id, should_exists=True)
        state = run.get_state()

        with run.fetch_scratch("request.json") as f:
            if f is None:
                request = {}
            else:
                request = json.load(f)

        cmd = run.fetch_state("cmd", "").split("\n")
        start_time = run.fetch_state("start_time")
        end_time = run.fetch_state("end_time")

        stdout = ""
        if run.get_stdout_path() is not None:
            # We have a standard output link.
            # We can't use flask_request.host_url here because that's just the
            # hostname, and we need to work mounted at a proxy hostname *and*
            # path under that hostname. So we need to use a relative URL to the
            # logs.
            stdout = f"../../../../toil/wes/v1/logs/{run_id}/stdout"
        stderr = ""
        if run.get_stderr_path() is not None:
            # We have a standard error link.
            stderr = f"../../../../toil/wes/v1/logs/{run_id}/stderr"

        exit_code = run.fetch_state("exit_code")

        if self.wes_dialect == "agc":
            # We need to emit a restricted subset of WES where tasks obey
            # certain Amazon-defined rules.
            logger.info("WES dialect %s requires transforming tasks", self.wes_dialect)
            filter_function = amazon_wes_utils.task_filter
        else:
            # We can emit any standard-compliant WES tasks
            logger.info(
                "WES dialect %s does not require transforming tasks", self.wes_dialect
            )
            filter_function = None
        task_logs = run.get_task_logs(filter_function=filter_function)

        output_obj = {}
        if state == "COMPLETE":
            output_obj = run.get_output_files()

        return {
            "run_id": run_id,
            "request": request,
            "state": state,
            "run_log": {
                "cmd": cmd,
                "start_time": start_time,
                "end_time": end_time,
                "stdout": stdout,
                "stderr": stderr,
                "exit_code": int(exit_code) if exit_code is not None else None,
            },
            "task_logs": task_logs,
            "outputs": output_obj,
        }

    @handle_errors
    def cancel_run(self, run_id: str) -> dict[str, str]:
        """Cancel a running workflow."""
        run = self._get_run(run_id, should_exists=True)

        # Do some preflight checks on the current state.
        # We won't catch all cases where the cancel won't go through, but we can catch some.
        state = run.get_state()
        if state in ("CANCELING", "CANCELED", "COMPLETE"):
            # We don't need to do anything.
            logger.warning(
                f"A user is attempting to cancel a workflow in state: '{state}'."
            )
        elif state in ("EXECUTOR_ERROR", "SYSTEM_ERROR"):
            # Something went wrong. Let the user know.
            raise OperationForbidden(
                f"Workflow is in state: '{state}', which cannot be cancelled."
            )
        else:
            # Go to canceling state if allowed
            run.state_machine.send_cancel()
            # Stop the run task if it is there.
            self.task_runner.cancel(run_id)

        return {"run_id": run_id}

    @handle_errors
    def get_run_status(self, run_id: str) -> dict[str, str]:
        """
        Get quick status info about a workflow run, returning a simple result
        with the overall state of the workflow run.
        """

        return {"run_id": run_id, "state": self.get_state(run_id)}

    # Toil custom endpoints that are not part of the GA4GH WES spec

    @handle_errors
    def get_stdout(self, run_id: str) -> Any:
        """
        Get the stdout of a workflow run as a static file.
        """
        run = self._get_run(run_id, should_exists=True)
        path = run.get_stdout_path()
        if path is None:
            # There's no log right now
            raise FileNotFoundError()
        else:
            return send_from_directory(run.scratch_dir, path, mimetype="text/plain")

    @handle_errors
    def get_stderr(self, run_id: str) -> Any:
        """
        Get the stderr of a workflow run as a static file.
        """
        run = self._get_run(run_id, should_exists=True)
        path = run.get_stderr_path()
        if path is None:
            # There's no log right now
            raise FileNotFoundError()
        else:
            return send_from_directory(run.scratch_dir, path, mimetype="text/plain")

    @handle_errors
    def get_health(self) -> Response:
        """
        Return successfully if the server is healthy.
        """
        return Response("OK", mimetype="text/plain")

    @handle_errors
    def get_homepage(self) -> Response:
        """
        Provide a sensible result for / other than 404.
        """
        # For now just go to the service info endpoint
        return redirect("ga4gh/wes/v1/service-info", code=302)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import ast
import json
import logging
import os
import pickle
import pwd
import socket
import time
import traceback
from argparse import ArgumentParser, _ArgumentGroup
from queue import Empty, Queue
from typing import Optional, Union
from urllib.parse import quote_plus
from urllib.request import urlopen

import addict
from pymesos import MesosSchedulerDriver, Scheduler, decode_data, encode_data

from toil import resolveEntryPoint
from toil.batchSystems.abstractBatchSystem import (
    EXIT_STATUS_UNAVAILABLE_VALUE,
    AbstractScalableBatchSystem,
    BatchJobExitReason,
    NodeInfo,
    UpdatedBatchJobInfo,
)
from toil.batchSystems.local_support import BatchSystemLocalSupport
from toil.batchSystems.mesos import JobQueue, MesosShape, TaskData, ToilJob
from toil.batchSystems.options import OptionSetter
from toil.common import Config
from toil.job import JobDescription
from toil.lib.conversions import b_to_mib, mib_to_b
from toil.lib.memoize import strict_bool
from toil.lib.misc import get_public_ip, get_user_name

log = logging.getLogger(__name__)


class MesosBatchSystem(BatchSystemLocalSupport, AbstractScalableBatchSystem, Scheduler):
    """
    A Toil batch system implementation that uses Apache Mesos to distribute toil jobs as Mesos
    tasks over a cluster of agent nodes. A Mesos framework consists of a scheduler and an
    executor. This class acts as the scheduler and is typically run on the master node that also
    runs the Mesos master process with which the scheduler communicates via a driver component.
    The executor is implemented in a separate class. It is run on each agent node and
    communicates with the Mesos agent process via another driver object. The scheduler may also
    be run on a separate node from the master, which we then call somewhat ambiguously the driver
    node.
    """

    @classmethod
    def supportsAutoDeployment(cls) -> bool:
        return True

    @classmethod
    def supportsWorkerCleanup(cls) -> bool:
        return True

    class ExecutorInfo:
        def __init__(
            self, nodeAddress: str, agentId: str, nodeInfo: str, lastSeen: str
        ) -> None:
            super().__init__()
            self.nodeAddress = nodeAddress
            self.agentId = agentId
            self.nodeInfo = nodeInfo
            self.lastSeen = lastSeen

    def __init__(
        self, config: Config, maxCores: float, maxMemory: float, maxDisk: int
    ) -> None:
        super().__init__(config, maxCores, maxMemory, maxDisk)

        # The auto-deployed resource representing the user script. Will be passed along in every
        # Mesos task. Also see setUserScript().
        self.userScript = None
        """
        :type: toil.resource.Resource
        """

        # Dictionary of queues, which toil assigns jobs to. Each queue represents a job type,
        # defined by resource usage
        self.jobQueues = JobQueue()

        # Address of the Mesos master in the form host:port where host can be an IP or a hostname
        self.mesos_endpoint = config.mesos_endpoint or self.get_default_mesos_endpoint()
        if config.mesos_role is not None:
            self.mesos_role = config.mesos_role
        self.mesos_name = config.mesos_name
        if config.mesos_framework_id is not None:
            self.mesos_framework_id = config.mesos_framework_id

        # Written to when Mesos kills tasks, as directed by Toil.
        # Jobs must not enter this set until they are removed from runningJobMap.
        self.killedJobIds = set()

        # The IDs of job to be killed
        self.killJobIds = set()

        # Contains jobs on which killBatchJobs were called, regardless of whether or not they
        # actually were killed or ended by themselves
        self.intendedKill = set()

        # Map of host address to job ids
        # this is somewhat redundant since Mesos returns the number of workers per
        # node. However, that information isn't guaranteed to reach the leader,
        # so we also track the state here. When the information is returned from
        # mesos, prefer that information over this attempt at state tracking.
        self.hostToJobIDs = {}

        # see self.setNodeFilter
        self.nodeFilter = []

        # Dict of launched jobIDs to TaskData objects
        self.runningJobMap = {}

        # Mesos has no easy way of getting a task's resources so we track them here
        self.taskResources = {}

        # Queue of jobs whose status has been updated, according to Mesos
        self.updatedJobsQueue = Queue()

        # The Mesos driver used by this scheduler
        self.driver = None

        # The string framework ID that we are assigned when registering with the Mesos master
        self.frameworkId = None

        # A dictionary mapping a node's IP to an ExecutorInfo object describing important
        # properties of our executor running on that node. Only an approximation of the truth.
        self.executors = {}

        # A dictionary mapping back from agent ID to the last observed IP address of its node.
        self.agentsByID = {}

        # A set of Mesos agent IDs, one for each agent running on a
        # non-preemptible node. Only an approximation of the truth. Recently
        # launched nodes may be absent from this set for a while and a node's
        # absence from this set does not imply its preemptability. But it is
        # generally safer to assume a node is preemptible since
        # non-preemptability is a stronger requirement. If we tracked the set
        # of preemptible nodes instead, we'd have to use absence as an
        # indicator of non-preemptability and could therefore be misled into
        # believing that a recently launched preemptible node was
        # non-preemptible.
        self.nonPreemptibleNodes = set()

        self.executor = self._buildExecutor()

        # These control how frequently to log a message that would indicate if no jobs are
        # currently able to run on the offers given. This can happen if the cluster is busy
        # or if the nodes in the cluster simply don't have enough resources to run the jobs
        self.lastTimeOfferLogged = 0
        self.logPeriod = 30  # seconds

        self.ignoredNodes = set()

        self._startDriver(config)

    def setUserScript(self, userScript: str) -> None:
        self.userScript = userScript

    def ignoreNode(self, nodeAddress: str) -> None:
        self.ignoredNodes.add(nodeAddress)

    def unignoreNode(self, nodeAddress: str) -> None:
        self.ignoredNodes.remove(nodeAddress)

    def issueBatchJob(
        self,
        command: str,
        jobNode: JobDescription,
        job_environment: Optional[dict[str, str]] = None,
    ) -> str:
        """
        Issues the following command returning a unique jobID. Command is the string to run, memory
        is an int giving the number of bytes the job needs to run in and cores is the number of cpus
        needed for the job and error-file is the path of the file to place any std-err/std-out in.
        """
        localID = self.handleLocalJob(command, jobNode)
        if localID is not None:
            return localID

        self.check_resource_request(jobNode)
        mesos_resources = {
            "memory": jobNode.memory,
            "cores": jobNode.cores,
            "disk": jobNode.disk,
            "preemptible": jobNode.preemptible,
        }

        jobID = self.getNextJobID()
        environment = self.environment.copy()
        if job_environment:
            environment.update(job_environment)

        job = ToilJob(
            jobID=jobID,
            name=str(jobNode),
            resources=MesosShape(wallTime=0, **mesos_resources),
            command=command,
            userScript=self.userScript,
            environment=environment,
            workerCleanupInfo=self.workerCleanupInfo,
        )
        jobType = job.resources
        log.debug("Queueing the job %s with job id: %s ...", jobNode, str(jobID))

        # TODO: round all elements of resources

        self.taskResources[jobID] = job.resources
        self.jobQueues.insertJob(job, jobType)
        log.debug("... queued")
        return jobID

    def killBatchJobs(self, jobIDs):

        # Some jobs may be local. Kill them first.
        self.killLocalJobs(jobIDs)

        # The driver thread does the actual work of killing the remote jobs.
        # We have to give it instructions, and block until the jobs are killed.
        assert self.driver is not None

        # This is the set of jobs that this invocation has asked to be killed,
        # but which haven't been killed yet.
        localSet = set()

        for jobID in jobIDs:
            # Queue the job up to be killed
            self.killJobIds.add(jobID)
            localSet.add(jobID)
            # Record that we meant to kill it, in case it finishes up by itself.
            self.intendedKill.add(jobID)

            if jobID in self.getIssuedBatchJobIDs():
                # Since the job has been issued, we have to kill it
                taskId = addict.Dict()
                taskId.value = str(jobID)
                log.debug("Kill issued job %s" % str(jobID))
                self.driver.killTask(taskId)
            else:
                # This job was never issued. Maybe it is a local job.
                # We don't have to kill it.
                log.debug("Skip non-issued job %s" % str(jobID))
                self.killJobIds.remove(jobID)
                localSet.remove(jobID)
        # Now localSet just has the non-local/issued jobs that we asked to kill
        while localSet:
            # Wait until they are all dead
            intersection = localSet.intersection(self.killedJobIds)
            if intersection:
                localSet -= intersection
                # When jobs are killed that we asked for, clear them out of
                # killedJobIds where the other thread put them
                self.killedJobIds -= intersection
            else:
                time.sleep(1)
        # Now all the jobs we asked to kill are dead. We know they are no
        # longer running, because that update happens before their IDs go into
        # killedJobIds. So we can safely return.

    def getIssuedBatchJobIDs(self):
        jobIds = set(self.jobQueues.jobIDs())
        jobIds.update(list(self.runningJobMap.keys()))
        return list(jobIds) + list(self.getIssuedLocalJobIDs())

    def getRunningBatchJobIDs(self):
        currentTime = dict()
        for jobID, data in list(self.runningJobMap.items()):
            currentTime[jobID] = time.time() - data.startTime
        currentTime.update(self.getRunningLocalJobIDs())
        return currentTime

    def getUpdatedBatchJob(self, maxWait):
        local_tuple = self.getUpdatedLocalJob(0)
        if local_tuple:
            return local_tuple
        while True:
            try:
                item = self.updatedJobsQueue.get(timeout=maxWait)
            except Empty:
                return None
            try:
                self.intendedKill.remove(item.jobID)
            except KeyError:
                log.debug(
                    "Job %s ended with status %i, took %s seconds.",
                    item.jobID,
                    item.exitStatus,
                    "???" if item.wallTime is None else str(item.wallTime),
                )
                return item
            else:
                log.debug(
                    "Job %s ended naturally before it could be killed.", item.jobID
                )

    def nodeInUse(self, nodeIP: str) -> bool:
        return nodeIP in self.hostToJobIDs

    def getWaitDuration(self):
        """
        Gets the period of time to wait (floating point, in seconds) between checking for
        missing/overlong jobs.
        """
        return 1

    def _buildExecutor(self):
        """
        Creates and returns an ExecutorInfo-shaped object representing our executor implementation.
        """
        # The executor program is installed as a setuptools entry point by setup.py
        info = addict.Dict()
        info.name = "toil"
        info.command.value = resolveEntryPoint("_toil_mesos_executor")
        info.executor_id.value = "toil-%i" % os.getpid()
        info.source = pwd.getpwuid(os.getuid()).pw_name
        return info

    def _startDriver(self, config):
        """
        The Mesos driver thread which handles the scheduler's communication with the Mesos master
        """
        framework = addict.Dict()
        framework.user = (
            get_user_name()
        )  # We must determine the user name ourselves with pymesos
        framework.name = config.mesos_name
        framework.principal = framework.name
        if config.mesos_role is not None:
            framework.roles = config.mesos_role
            framework.capabilities = [dict(type="MULTI_ROLE")]

        # Make the driver which implements most of the scheduler logic and calls back to us for the user-defined parts.
        # Make sure it will call us with nice namespace-y addicts
        self.driver = MesosSchedulerDriver(
            self,
            framework,
            self._resolveAddress(self.mesos_endpoint),
            use_addict=True,
            implicit_acknowledgements=True,
        )
        self.driver.start()

    @staticmethod
    def _resolveAddress(address):
        """
        Resolves the host in the given string. The input is of the form host[:port]. This method
        is idempotent, i.e. the host may already be a dotted IP address.

        >>> # noinspection PyProtectedMember
        >>> f=MesosBatchSystem._resolveAddress
        >>> f('localhost')
        '127.0.0.1'
        >>> f('127.0.0.1')
        '127.0.0.1'
        >>> f('localhost:123')
        '127.0.0.1:123'
        >>> f('127.0.0.1:123')
        '127.0.0.1:123'
        """
        address = address.split(":")
        assert len(address) in (1, 2)
        address[0] = socket.gethostbyname(address[0])
        return ":".join(address)

    def shutdown(self) -> None:
        self.shutdownLocal()
        log.debug("Stopping Mesos driver")
        self.driver.stop()
        log.debug("Joining Mesos driver")
        driver_result = self.driver.join()
        log.debug("Joined Mesos driver")
        if driver_result is not None and driver_result != "DRIVER_STOPPED":
            # TODO: The docs say join should return a code, but it keeps returning
            # None when apparently successful. So tolerate that here too.
            raise RuntimeError("Mesos driver failed with %s" % driver_result)

    def registered(self, driver, frameworkId, masterInfo):
        """
        Invoked when the scheduler successfully registers with a Mesos master
        """
        log.debug("Registered with framework ID %s", frameworkId.value)
        # Save the framework ID
        self.frameworkId = frameworkId.value

    def _declineAllOffers(self, driver, offers):
        for offer in offers:
            driver.declineOffer(offer.id)

    def _parseOffer(self, offer):
        cores = 0
        memory = 0
        disk = 0
        preemptible = None
        for attribute in offer.attributes:
            if attribute.name == "preemptible":
                assert (
                    preemptible is None
                ), "Attribute 'preemptible' occurs more than once."
                preemptible = strict_bool(attribute.text.value)
        if preemptible is None:
            log.debug(
                "Agent not marked as either preemptible or not. Assuming non-preemptible."
            )
            preemptible = False
        for resource in offer.resources:
            if resource.name == "cpus":
                cores += resource.scalar.value
            elif resource.name == "mem":
                memory += resource.scalar.value
            elif resource.name == "disk":
                disk += resource.scalar.value
        return cores, memory, disk, preemptible

    def _prepareToRun(self, jobType, offer):
        # Get the first element to ensure FIFO
        job = self.jobQueues.nextJobOfType(jobType)
        task = self._newMesosTask(job, offer)
        return task

    def _updateStateToRunning(self, offer, runnableTasks):
        for task in runnableTasks:
            resourceKey = int(task.task_id.value)
            resources = self.taskResources[resourceKey]
            agentIP = socket.gethostbyname(offer.hostname)
            try:
                self.hostToJobIDs[agentIP].append(resourceKey)
            except KeyError:
                self.hostToJobIDs[agentIP] = [resourceKey]

            self.runningJobMap[int(task.task_id.value)] = TaskData(
                startTime=time.time(),
                agentID=offer.agent_id.value,
                agentIP=agentIP,
                executorID=task.executor.executor_id.value,
                cores=resources.cores,
                memory=resources.memory,
            )
            del self.taskResources[resourceKey]
            log.debug("Launched Mesos task %s.", task.task_id.value)

    def resourceOffers(self, driver, offers):
        """
        Invoked when resources have been offered to this framework.
        """
        self._trackOfferedNodes(offers)

        jobTypes = self.jobQueues.sortedTypes

        if not jobTypes:
            # Without jobs, we can get stuck with no jobs and no new offers until we decline it.
            self._declineAllOffers(driver, offers)
            return

        unableToRun = True
        # Right now, gives priority to largest jobs
        for offer in offers:
            if offer.hostname in self.ignoredNodes:
                driver.declineOffer(offer.id)
                continue
            runnableTasks = []
            # TODO: In an offer, can there ever be more than one resource with the same name?
            offerCores, offerMemory, offerDisk, offerPreemptible = self._parseOffer(
                offer
            )
            log.debug(
                "Got offer %s for a %spreemptible agent with %.2f MiB memory, %.2f core(s) "
                "and %.2f MiB of disk.",
                offer.id.value,
                "" if offerPreemptible else "non-",
                offerMemory,
                offerCores,
                offerDisk,
            )
            remainingCores = offerCores
            remainingMemory = offerMemory
            remainingDisk = offerDisk

            for jobType in jobTypes:
                runnableTasksOfType = []
                # Because we are not removing from the list until outside of the while loop, we
                # must decrement the number of jobs left to run ourselves to avoid an infinite
                # loop.
                nextToLaunchIndex = 0
                # Toil specifies disk and memory in bytes but Mesos uses MiB
                while (
                    not self.jobQueues.typeEmpty(jobType)
                    # On a non-preemptible node we can run any job, on a preemptible node we
                    # can only run preemptible jobs:
                    and (not offerPreemptible or jobType.preemptible)
                    and remainingCores >= jobType.cores
                    and remainingDisk >= b_to_mib(jobType.disk)
                    and remainingMemory >= b_to_mib(jobType.memory)
                ):
                    task = self._prepareToRun(jobType, offer)
                    # TODO: this used to be a conditional but Hannes wanted it changed to an assert
                    # TODO: ... so we can understand why it exists.
                    assert int(task.task_id.value) not in self.runningJobMap
                    runnableTasksOfType.append(task)
                    log.debug(
                        "Preparing to launch Mesos task %s with %.2f cores, %.2f MiB memory, and %.2f MiB disk using offer %s ...",
                        task.task_id.value,
                        jobType.cores,
                        b_to_mib(jobType.memory),
                        b_to_mib(jobType.disk),
                        offer.id.value,
                    )
                    remainingCores -= jobType.cores
                    remainingMemory -= b_to_mib(jobType.memory)
                    remainingDisk -= b_to_mib(jobType.disk)
                    nextToLaunchIndex += 1
                if not self.jobQueues.typeEmpty(jobType):
                    # report that remaining jobs cannot be run with the current resourcesq:
                    log.debug(
                        "Offer %(offer)s not suitable to run the tasks with requirements "
                        "%(requirements)r. Mesos offered %(memory)s memory, %(cores)s cores "
                        "and %(disk)s of disk on a %(non)spreemptible agent.",
                        dict(
                            offer=offer.id.value,
                            requirements=jobType.__dict__,
                            non="" if offerPreemptible else "non-",
                            memory=mib_to_b(offerMemory),
                            cores=offerCores,
                            disk=mib_to_b(offerDisk),
                        ),
                    )
                runnableTasks.extend(runnableTasksOfType)
            # Launch all runnable tasks together so we only call launchTasks once per offer
            if runnableTasks:
                unableToRun = False
                driver.launchTasks(offer.id, runnableTasks)
                self._updateStateToRunning(offer, runnableTasks)
            else:
                log.debug(
                    "Although there are queued jobs, none of them could be run with offer %s "
                    "extended to the framework.",
                    offer.id,
                )
                driver.declineOffer(offer.id)

        if unableToRun and time.time() > (self.lastTimeOfferLogged + self.logPeriod):
            self.lastTimeOfferLogged = time.time()
            log.debug(
                "Although there are queued jobs, none of them were able to run in "
                "any of the offers extended to the framework. There are currently "
                "%i jobs running. Enable debug level logging to see more details about "
                "job types and offers received.",
                len(self.runningJobMap),
            )

    def _trackOfferedNodes(self, offers):
        for offer in offers:
            # All AgentID messages are required to have a value according to the Mesos Protobuf file.
            assert "value" in offer.agent_id
            try:
                nodeAddress = socket.gethostbyname(offer.hostname)
            except:
                log.debug("Failed to resolve hostname %s" % offer.hostname)
                raise
            self._registerNode(nodeAddress, offer.agent_id.value)
            preemptible = False
            for attribute in offer.attributes:
                if attribute.name == "preemptible":
                    preemptible = strict_bool(attribute.text.value)
            if preemptible:
                try:
                    self.nonPreemptibleNodes.remove(offer.agent_id.value)
                except KeyError:
                    pass
            else:
                self.nonPreemptibleNodes.add(offer.agent_id.value)

    def _filterOfferedNodes(self, offers):
        if not self.nodeFilter:
            return offers
        executorInfoOrNone = [
            self.executors.get(socket.gethostbyname(offer.hostname)) for offer in offers
        ]
        executorInfos = [_f for _f in executorInfoOrNone if _f]
        executorsToConsider = list(filter(self.nodeFilter[0], executorInfos))
        ipsToConsider = {ex.nodeAddress for ex in executorsToConsider}
        return [
            offer
            for offer in offers
            if socket.gethostbyname(offer.hostname) in ipsToConsider
        ]

    def _newMesosTask(self, job, offer):
        """
        Build the Mesos task object for a given the Toil job and Mesos offer
        """
        task = addict.Dict()
        task.task_id.value = str(job.jobID)
        task.agent_id.value = offer.agent_id.value
        task.name = job.name
        task.data = encode_data(pickle.dumps(job))
        task.executor = addict.Dict(self.executor)

        task.resources = []

        task.resources.append(addict.Dict())
        cpus = task.resources[-1]
        cpus.name = "cpus"
        cpus.type = "SCALAR"
        cpus.scalar.value = job.resources.cores

        task.resources.append(addict.Dict())
        disk = task.resources[-1]
        disk.name = "disk"
        disk.type = "SCALAR"
        if b_to_mib(job.resources.disk) > 1:
            disk.scalar.value = b_to_mib(job.resources.disk)
        else:
            log.warning(
                "Job %s uses less disk than Mesos requires. Rounding %s up to 1 MiB.",
                job.jobID,
                job.resources.disk,
            )
            disk.scalar.value = 1

        task.resources.append(addict.Dict())
        mem = task.resources[-1]
        mem.name = "mem"
        mem.type = "SCALAR"
        if b_to_mib(job.resources.memory) > 1:
            mem.scalar.value = b_to_mib(job.resources.memory)
        else:
            log.warning(
                "Job %s uses less memory than Mesos requires. Rounding %s up to 1 MiB.",
                job.jobID,
                job.resources.memory,
            )
            mem.scalar.value = 1
        return task

    def statusUpdate(self, driver, update):
        """
        Invoked when the status of a task has changed (e.g., a agent is lost and so the task is
        lost, a task finishes and an executor sends a status update saying so, etc). Note that
        returning from this callback _acknowledges_ receipt of this status update! If for
        whatever reason the scheduler aborts during this callback (or the process exits) another
        status update will be delivered (note, however, that this is currently not true if the
        agent sending the status update is lost/fails during that time).
        """
        jobID = int(update.task_id.value)
        log.debug(
            "Job %i is in state '%s' due to reason '%s'.",
            jobID,
            update.state,
            update.reason,
        )

        def jobEnded(_exitStatus, wallTime=None, exitReason=None):
            """
            Notify external observers of the job ending.
            """
            self.updatedJobsQueue.put(
                UpdatedBatchJobInfo(
                    jobID=jobID,
                    exitStatus=_exitStatus,
                    wallTime=wallTime,
                    exitReason=exitReason,
                )
            )
            agentIP = None
            try:
                agentIP = self.runningJobMap[jobID].agentIP
            except KeyError:
                log.warning(
                    "Job %i returned exit code %i but isn't tracked as running.",
                    jobID,
                    _exitStatus,
                )
            else:
                # Mark the job as no longer running. We MUST do this BEFORE
                # saying we killed the job, or it will be possible for another
                # thread to kill a job and then see it as running.
                del self.runningJobMap[jobID]

            try:
                self.hostToJobIDs[agentIP].remove(jobID)
            except KeyError:
                log.warning(
                    "Job %i returned exit code %i from unknown host.",
                    jobID,
                    _exitStatus,
                )

            try:
                self.killJobIds.remove(jobID)
            except KeyError:
                pass
            else:
                # We were asked to kill this job, so say that we have done so.
                # We do this LAST, after all status updates for the job have
                # been handled, to ensure a consistent view of the scheduler
                # state from other threads.
                self.killedJobIds.add(jobID)

        if update.state == "TASK_FINISHED":
            # We get the running time of the job via the timestamp, which is in job-local time in seconds
            labels = update.labels.labels
            wallTime = None
            for label in labels:
                if label["key"] == "wallTime":
                    wallTime = float(label["value"])
                    break
            assert wallTime is not None
            jobEnded(0, wallTime=wallTime, exitReason=BatchJobExitReason.FINISHED)
        elif update.state == "TASK_FAILED":
            try:
                exitStatus = int(update.message)
            except ValueError:
                exitStatus = EXIT_STATUS_UNAVAILABLE_VALUE
                log.warning(
                    "Job %i failed with message '%s' due to reason '%s' on executor '%s' on agent '%s'.",
                    jobID,
                    update.message,
                    update.reason,
                    update.executor_id,
                    update.agent_id,
                )
            else:
                log.warning(
                    "Job %i failed with exit status %i and message '%s' due to reason '%s' on executor '%s' on agent '%s'.",
                    jobID,
                    exitStatus,
                    update.message,
                    update.reason,
                    update.executor_id,
                    update.agent_id,
                )

            jobEnded(exitStatus, exitReason=BatchJobExitReason.FAILED)
        elif update.state == "TASK_LOST":
            log.warning("Job %i is lost.", jobID)
            jobEnded(EXIT_STATUS_UNAVAILABLE_VALUE, exitReason=BatchJobExitReason.LOST)
        elif update.state in ("TASK_KILLED", "TASK_ERROR"):
            log.warning(
                "Job %i is in unexpected state %s with message '%s' due to reason '%s'.",
                jobID,
                update.state,
                update.message,
                update.reason,
            )
            jobEnded(
                EXIT_STATUS_UNAVAILABLE_VALUE,
                exitReason=(
                    BatchJobExitReason.KILLED
                    if update.state == "TASK_KILLED"
                    else BatchJobExitReason.ERROR
                ),
            )

        if "limitation" in update:
            log.warning("Job limit info: %s" % update.limitation)

    def frameworkMessage(self, driver, executorId, agentId, message):
        """
        Invoked when an executor sends a message.
        """

        # Take it out of base 64 encoding from Protobuf
        message = decode_data(message).decode()

        log.debug(
            "Got framework message from executor %s running on agent %s: %s",
            executorId.value,
            agentId.value,
            message,
        )
        message = ast.literal_eval(message)
        assert isinstance(message, dict)
        # Handle the mandatory fields of a message
        nodeAddress = message.pop("address")
        executor = self._registerNode(nodeAddress, agentId.value)
        # Handle optional message fields
        for k, v in message.items():
            if k == "nodeInfo":
                assert isinstance(v, dict)
                resources = [
                    taskData
                    for taskData in self.runningJobMap.values()
                    if taskData.executorID == executorId.value
                ]
                requestedCores = sum(taskData.cores for taskData in resources)
                requestedMemory = sum(taskData.memory for taskData in resources)
                executor.nodeInfo = NodeInfo(
                    requestedCores=requestedCores, requestedMemory=requestedMemory, **v
                )
                self.executors[nodeAddress] = executor
            else:
                raise RuntimeError("Unknown message field '%s'." % k)

    def _registerNode(self, nodeAddress, agentId, nodePort=5051):
        """
        Called when we get communication from an agent. Remembers the
        information about the agent by address, and the agent address by agent
        ID.
        """
        executor = self.executors.get(nodeAddress)
        if executor is None or executor.agentId != agentId:
            executor = self.ExecutorInfo(
                nodeAddress=nodeAddress,
                agentId=agentId,
                nodeInfo=None,
                lastSeen=time.time(),
            )
            self.executors[nodeAddress] = executor
        else:
            executor.lastSeen = time.time()

        # Record the IP under the agent id
        self.agentsByID[agentId] = nodeAddress

        return executor

    def getNodes(
        self, preemptible: Optional[bool] = None, timeout: Optional[int] = None
    ) -> dict[str, NodeInfo]:
        """
        Return all nodes that match:
         - preemptible status (None includes all)
         - timeout period (seen within the last # seconds, or None for all)
        """
        nodes = dict()
        for node_ip, executor in self.executors.items():
            if preemptible is None or (
                preemptible == (executor.agentId not in self.nonPreemptibleNodes)
            ):
                if timeout is None or (time.time() - executor.lastSeen < timeout):
                    nodes[node_ip] = executor.nodeInfo
        return nodes

    def reregistered(self, driver, masterInfo):
        """
        Invoked when the scheduler re-registers with a newly elected Mesos master.
        """
        log.debug("Registered with new master")

    def _handleFailedExecutor(self, agentID, executorID=None):
        """
        Should be called when we find out an executor has failed.

        Gets the log from some container (since we are never handed a container
        ID) that ran on the given executor on the given agent, if the agent is
        still up, and dumps it to our log. All IDs are strings.

        If executorID is None, dumps all executors from the agent.

        Useful for debugging failing executor code.
        """

        log.warning(
            "Handling failure of executor '%s' on agent '%s'.", executorID, agentID
        )

        try:
            # Look up the IP. We should always know it unless we get answers
            # back without having accepted offers.
            agentAddress = self.agentsByID[agentID]

            # For now we assume the agent is always on the same port. We could
            # maybe sniff this from the URL that comes in the offer but it's
            # not guaranteed to be there.
            agentPort = 5051

            # We need the container ID to read the log, but we are never given
            # it, and I can't find a good way to list it, because the API only
            # seems to report running containers. So we dump all the available
            # files with /files/debug and look for one that looks right.
            filesQueryURL = errorLogURL = "http://%s:%d/files/debug" % (
                agentAddress,
                agentPort,
            )

            # Download all the root mount points, which are in an object from
            # mounted name to real name
            filesDict = json.loads(urlopen(filesQueryURL).read())

            log.debug("Available files: %s", repr(filesDict.keys()))

            # Generate filenames for each container pointing to where stderr should be
            stderrFilenames = []
            # And look for the actual agent logs.
            agentLogFilenames = []
            for filename in filesDict:
                if (
                    self.frameworkId in filename
                    and agentID in filename
                    and (executorID is None or executorID in filename)
                ):

                    stderrFilenames.append("%s/stderr" % filename)
                elif filename.endswith("log"):
                    agentLogFilenames.append(filename)

            if len(stderrFilenames) == 0:
                log.warning("Could not find any containers in '%s'." % filesDict)

            for stderrFilename in stderrFilenames:
                try:

                    # According to
                    # http://mesos.apache.org/documentation/latest/sandbox/ we can use
                    # the web API to fetch the error log.
                    errorLogURL = "http://%s:%d/files/download?path=%s" % (
                        agentAddress,
                        agentPort,
                        quote_plus(stderrFilename),
                    )

                    log.warning(
                        "Attempting to retrieve executor error log: %s", errorLogURL
                    )

                    for line in urlopen(errorLogURL):
                        # Warn all the lines of the executor's error log
                        log.warning("Executor: %s", line.rstrip())

                except Exception as e:
                    log.warning("Could not retrieve exceutor log due to: '%s'.", e)
                    log.warning(traceback.format_exc())

            for agentLogFilename in agentLogFilenames:
                try:
                    agentLogURL = "http://%s:%d/files/download?path=%s" % (
                        agentAddress,
                        agentPort,
                        quote_plus(agentLogFilename),
                    )

                    log.warning("Attempting to retrieve agent log: %s", agentLogURL)

                    for line in urlopen(agentLogURL):
                        # Warn all the lines of the agent's log
                        log.warning("Agent: %s", line.rstrip())
                except Exception as e:
                    log.warning("Could not retrieve agent log due to: '%s'.", e)
                    log.warning(traceback.format_exc())

        except Exception as e:
            log.warning("Could not retrieve logs due to: '%s'.", e)
            log.warning(traceback.format_exc())

    def executorLost(self, driver, executorId, agentId, status):
        """
        Invoked when an executor has exited/terminated abnormally.
        """

        failedId = executorId.get("value", None)

        log.warning("Executor '%s' reported lost with status '%s'.", failedId, status)

        self._handleFailedExecutor(agentId.value, failedId)

    @classmethod
    def get_default_mesos_endpoint(cls) -> str:
        """
        Get the default IP/hostname and port that we will look for Mesos at.
        """
        return f"{get_public_ip()}:5050"

    @classmethod
    def add_options(cls, parser: Union[ArgumentParser, _ArgumentGroup]) -> None:
        parser.add_argument(
            "--mesosEndpoint",
            "--mesosMaster",
            dest="mesos_endpoint",
            default=None,
            help=f"The host and port of the Mesos master separated by colon. If the provided value "
            f"is None, the value will be generated at runtime.  "
            f"(Generated default: {cls.get_default_mesos_endpoint})",
        )
        parser.add_argument(
            "--mesosFrameworkId",
            dest="mesos_framework_id",
            help="Use a specific Mesos framework ID.",
        )
        parser.add_argument("--mesosRole", dest="mesos_role", help="Use a Mesos role.")
        parser.add_argument(
            "--mesosName",
            dest="mesos_name",
            default="toil",
            help="The Mesos name to use. (default: %(default)s)",
        )

    @classmethod
    def setOptions(cls, setOption: OptionSetter):
        setOption("mesos_endpoint", old_names=["mesosMasterAddress"])
        setOption("mesos_name")
        setOption("mesos_role")
        setOption("mesos_framework_id")



# Copyright (C) 2015-2022 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import getpass
import logging

from toil.lib.misc import get_user_name
from toil.test import ToilTest

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


class UserNameAvailableTest(ToilTest):
    """
    Make sure we can get user names when they are available.
    """

    def test_get_user_name(self):
        # We assume we have the user in /etc/passwd when running the tests.
        real_user_name = getpass.getuser()
        apparent_user_name = get_user_name()
        self.assertEqual(apparent_user_name, real_user_name)


class UserNameUnvailableTest(ToilTest):
    """
    Make sure we can get something for a user name when user names are not
    available.
    """

    def setUp(self):
        super().setUp()
        # Monkey patch getpass.getuser to fail
        self.original_getuser = getpass.getuser

        def fake_getuser():
            raise KeyError("Fake key error")

        getpass.getuser = fake_getuser

    def tearDown(self):
        # Fix the module we hacked up
        getpass.getuser = self.original_getuser
        super().tearDown()

    def test_get_user_name(self):
        apparent_user_name = get_user_name()
        # Make sure we got something
        self.assertTrue(isinstance(apparent_user_name, str))
        self.assertNotEqual(apparent_user_name, "")


class UserNameVeryBrokenTest(ToilTest):
    """
    Make sure we can get something for a user name when user name fetching is
    broken in ways we did not expect.
    """

    def setUp(self):
        super().setUp()
        # Monkey patch getpass.getuser to fail
        self.original_getuser = getpass.getuser

        def fake_getuser():
            raise RuntimeError("Fake error that we did not anticipate")

        getpass.getuser = fake_getuser

    def tearDown(self):
        # Fix the module we hacked up
        getpass.getuser = self.original_getuser
        super().tearDown()

    def test_get_user_name(self):
        apparent_user_name = get_user_name()
        # Make sure we got something
        self.assertTrue(isinstance(apparent_user_name, str))
        self.assertNotEqual(apparent_user_name, "")



baseVersion = '8.2.0'
cgcloudVersion = '1.6.0a1.dev393'
version = '8.2.0-3efbe6f84b89a1eeab1d7b5913bbd01084d6490e'
cacheTag = 'cache-local-py3.10'
mainCacheTag = 'cache-master-py3.10'
distVersion = '8.2.0'
exactPython = 'python3.10'
python = 'python3.10'
dockerTag = '8.2.0-3efbe6f84b89a1eeab1d7b5913bbd01084d6490e-py3.10'
currentCommit = '3efbe6f84b89a1eeab1d7b5913bbd01084d6490e'
dockerRegistry = 'quay.io/ucsc_cgl'
dockerName = 'toil'
dirty = False
cwltool_version = '3.1.20250110105449'



#!/usr/bin/env python
"""
cleanup_aws_resources.py

Convenience script to clean up extraneous test buckets, sbd domains, instance profiles, and/or roles
that testing may have left behind.

Manually requires the user to inspect and submit "yes" or "y" to delete buckets/domains.

Failing or canceled tests may miss clean up, and leftover test buckets and sdb domains can build up.
Run this script occasionally or when needed (there are limits to the number of buckets/domains we can have and we
can hit those limits).
"""
import argparse
import copy
import os
import re
import sys

pkg_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))  # noqa
sys.path.insert(0, pkg_root)  # noqa

from src.toil.lib import aws
from src.toil.lib.aws import session
from src.toil.lib.aws.iam import delete_iam_instance_profile, delete_iam_role
from src.toil.lib.aws.utils import delete_s3_bucket, delete_sdb_domain
from src.toil.lib.generatedEC2Lists import regionDict

# put us-west-2 first as our default test region; that way anything with a universal region shows there
regions = ['us-west-2'] + [region for region in regionDict if region != 'us-west-2']

# never show these buckets; never offer to delete them; never forget
absolutely_do_not_delete_these_buckets = ['318423852362-cgcloud',  # not sure what this is
                                          'aws-config20201211232942693800000001',  # AWS logging; ask Erich?
                                          'cgl-pipeline',  # something important?
                                          'cgl-rnaseq-recompute-fixed-toil',  # the 20,000 toil-rna-seq recompute data
                                          'toil-cloudtrail-bucket',  # also AWS logging... ?; ask Erich
                                          'toil-cwl-infra-test-bucket-dont-delete',  # test infra; never delete
                                          'toil-datasets',  # test infra; never delete
                                          'toil-no-location-bucket-dont-delete',  # test infra; never delete
                                          'toil-preserve-file-permissions-tests']  # test infra; never delete


def contains_uuid(string):
    """
    Determines if a string contains a pattern like: '28064c76-a491-43e7-9b50-da424f920354',
    which toil uses in its test generated bucket names.
    """
    return bool(re.compile('[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{8,12}').findall(string))


def contains_uuid_with_underscores(string):
    """
    Determines if a string contains a pattern like: '28064c76-a491-43e7-9b50-da424f920354',
    which toil uses in its test generated IAM role names.
    """
    return bool(re.compile('[0-9a-f]{8}_[0-9a-f]{4}_[0-9a-f]{4}_[0-9a-f]{4}_[0-9a-f]{12}').findall(string))


def contains_num_only_uuid(string):
    """
    Determines if a string contains a pattern like: '13614-31311-31347',
    which toil uses in its test generated sdb domain names.
    """
    return bool(re.compile('[0-9]{4,5}-[0-9]{4,5}-[0-9]{4,5}').findall(string))


def contains_toil_test_patterns(string):
    return contains_uuid(string) or contains_num_only_uuid(string) or contains_uuid_with_underscores(string)


def matches(resource_name):
    if (resource_name.endswith('--files') or resource_name.endswith('--jobs') or resource_name.endswith('_toil')
            or resource_name.endswith('--internal') or resource_name.startswith('toil-s3test-')):
        if contains_toil_test_patterns(resource_name):
            return resource_name

    if resource_name.startswith('import-export-test-'):
        return resource_name


def find_buckets_to_cleanup(include_all, match):
    buckets = dict()
    for region in regions:
        print(f'\n[{region}] Buckets:')
        try:
            s3_resource = session.resource('s3', region_name=region)
            buckets_in_region = find_buckets_in_region(s3_resource, include_all, match)
            new_buckets = [b for b in buckets_in_region if b not in buckets]
            print('    ' + '\n    '.join(new_buckets))
            for bucket in new_buckets:
                buckets[bucket] = region
        except Exception as e:
            # Occurs with botocore.exceptions.ClientError.
            if 'Your account is not signed up for the S3 service' in str(e):
                print('    Your account is not signed up for the S3 service in this region.')
            else:
                print(f'    An error occurred in this region: {e}')
    return buckets


def find_sdb_domains_to_cleanup(include_all, match):
    sdb_domains = dict()
    for region in regions:
        print(f'\n[{region}] SimpleDB Domains:')
        try:
            sdb_client = session.client('sdb', region_name=region)
            domains_in_region = find_sdb_domains_in_region(sdb_client, include_all, match)
            new_domains = [b for b in domains_in_region if b not in sdb_domains]
            print('    ' + '\n    '.join(new_domains))
            for sdb_domain in new_domains:
                sdb_domains[sdb_domain] = region
        except Exception as e:
            # Occurs with botocore.exceptions.SSLError in regions that don't support SDB.
            # Don't hard-code supported regions, just in case AWS changes these, in order to avoid blind spots.
            if 'SSL validation failed' in str(e):
                print('    SimpleDB is not offered in this region.')
            else:
                print(f'    An error occurred in this region: {e}')
    return sdb_domains


def find_iam_roles_to_cleanup(include_all, match):
    iam_roles = dict()
    for region in regions:
        print(f'\n[{region}] IAM Roles:')
        try:
            iam_client = session.client('iam', region_name=region)
            roles_in_region = find_iam_roles_in_region(iam_client, include_all, match)

            new_roles = [b for b in roles_in_region if b not in iam_roles]
            print('    ' + '\n    '.join(new_roles))
            for iam_role in new_roles:
                iam_roles[iam_role] = region
        except Exception as e:
            print(f'    An error occurred in this region: {e}')
    return iam_roles


def find_instance_profile_names_to_cleanup(include_all, match):
    instance_profiles = dict()
    for region in regions:
        print(f'\n[{region}] IAM Instance Profiles:')
        try:
            iam_resource = session.resource('iam', region_name=region)
            iam_client = session.client('iam')
            instance_profiles_in_region = find_instance_profile_names_in_region(iam_client, include_all, match)

            new_instance_profiles = [b for b in instance_profiles_in_region if b not in instance_profiles]
            for new_profile in new_instance_profiles:
                print(f'    {new_profile}')
                associated_roles = [r.name for r in iam_resource.InstanceProfile(new_profile).roles]
                print('     - Roles:')
                if associated_roles:
                    print('       - ' + f'\n       - '.join(associated_roles) + '\n')
                else:
                    print('       - No roles found to be associated.\n')
                instance_profiles[new_profile] = region
        except Exception as e:
            print(f'    An error occurred in this region: {e}')
    return instance_profiles


def find_buckets_in_region(s3_resource, include_all, match):
    buckets_to_cleanup = []
    for bucket in s3_resource.buckets.all():
        if bucket.name not in absolutely_do_not_delete_these_buckets:
            if match:
                for m in match:
                    if m in bucket.name:
                        buckets_to_cleanup.append(bucket.name)
            elif matches(bucket.name) or include_all:
                buckets_to_cleanup.append(bucket.name)
    return buckets_to_cleanup


def find_sdb_domains_in_region(sdb_client, include_all, match):
    sdb_domains_to_cleanup = []
    paginator = sdb_client.get_paginator('list_domains')
    for sdb_domains in paginator.paginate():
        for sdb_domain in sdb_domains.get('DomainNames', []):
            if match:
                for m in match:
                    if m in sdb_domain:
                        sdb_domains_to_cleanup.append(sdb_domain)
            elif matches(sdb_domain) or include_all:
                sdb_domains_to_cleanup.append(sdb_domain)
    return sdb_domains_to_cleanup


def find_iam_roles_in_region(iam_client, include_all, match):
    iam_roles_to_cleanup = []
    paginator = iam_client.get_paginator('list_roles')
    for iam_roles in paginator.paginate():
        for iam_role in iam_roles.get('Roles', []):
            iam_role = iam_role['RoleName']
            if match:
                for m in match:
                    if m in iam_role:
                        iam_roles_to_cleanup.append(iam_role)
            elif matches(iam_role) or include_all:
                iam_roles_to_cleanup.append(iam_role)
    return iam_roles_to_cleanup


def find_instance_profile_names_in_region(iam_client, include_all, match):
    instance_profiles_to_cleanup = []
    paginator = iam_client.get_paginator('list_instance_profiles')
    for instance_profiles in paginator.paginate():
        for instance_profile in instance_profiles.get('InstanceProfiles', []):
            instance_profile = instance_profile['InstanceProfileName']
            if match:
                for m in match:
                    if m in instance_profile:
                        instance_profiles_to_cleanup.append(instance_profile)
            elif matches(instance_profile) or include_all:
                instance_profiles_to_cleanup.append(instance_profile)
    return instance_profiles_to_cleanup


def main(argv):
    parser = argparse.ArgumentParser(
        description='View and/or clean up s3 buckets and/or sdb domains in an AWS Account.')

    parser.add_argument("--include-all", dest='include_all',
                        action='store_true', required=False,
                        help="Don't filter based on buckets/domains that look like test objects.  "
                             "List everything in the account.")
    parser.add_argument("--view-only", dest='view_only',
                        action='store_true', required=False,
                        help="Don't ask to delete.  Just view everything.")
    parser.add_argument("--skip-buckets", dest='skip_buckets',
                        action='store_true', required=False,
                        help="Skip doing anything with buckets.")
    parser.add_argument("--skip-sdb", dest='skip_sdb',
                        action='store_true', required=False,
                        help="Skip doing anything with SimpleDB domains.")
    parser.add_argument("--skip-iam-roles", dest='skip_iam_roles',
                        action='store_true', required=False,
                        help="Skip doing anything with IAM roles.")
    parser.add_argument("--skip-iam-instance-profiles", dest='skip_iam_instance_profiles',
                        action='store_true', required=False,
                        help="Skip doing anything with IAM roles.")
    parser.add_argument("--match", dest='match',
                        type=str, required=False, default='',
                        help="Only return resources containing the comma-delimited keywords.  "
                             "For example, adding --match='hello,goodbye' would return any "
                             "buckets or domains that include either 'hello' or 'goodbye'.")
    parser.add_argument("--regions", dest='regions',
                        type=str, required=False, default='all',
                        help="Only return resources in the regions (comma-delimited).  "
                             "For example, adding --regions='us-west-2,us-east-1' will "
                             "only act resources in those two regions.")
    # parser.set_defaults(view_only=False, include_all=False, skip_buckets=False, skip_sdb=False, match='')

    options = parser.parse_args(argv)

    account_aliases = session.client('iam').list_account_aliases()['AccountAliases']
    account_name = account_aliases[0] if account_aliases else "[no name]"
    print(f'\n\nNow running for AWS account: {account_name}.')

    match = [m.strip() for m in options.match.split(',') if m.strip()]

    if match and options.include_all:
        raise ValueError('Cannot filter on match patterns AND include everything.  Please specify either '
                         '"--view-only" or "--match", but not both.')

    if options.regions and options.regions != 'all':
        options.regions = [r.strip() for r in options.regions.split(',') if r.strip()]
        frozen_regions = copy.deepcopy(regions)
        for region in frozen_regions:
            if region not in options.regions and region in regions:
                regions.remove(region)

    if not options.skip_buckets:
        buckets = find_buckets_to_cleanup(options.include_all, match)
        if not options.view_only:
            if not buckets:
                print('Nothing to be done.')
            else:
                response = input(f'\nDo you wish to delete these buckets in account: {account_name}?  (Y)es (N)o: ')
                if response.lower() in ('y', 'yes'):
                    print('\nOkay, now deleting...')
                    for bucket, region in buckets.items():
                        s3_resource = session.resource('s3', region_name=region)
                        delete_s3_bucket(s3_resource, bucket)
                    print('S3 Bucket Deletions Successful.')

    if not options.skip_sdb:
        sdb_domains = find_sdb_domains_to_cleanup(options.include_all, match)
        if not options.view_only:
            if not sdb_domains:
                print('Nothing to be done.')
            else:
                response = input(f'\nDo you wish to delete these SDB domains in account: {account_name}?  (Y)es (N)o: ')
                if response.lower() in ('y', 'yes'):
                    print('\nOkay, now deleting...')
                    for sdb_domain, region in sdb_domains.items():
                        delete_sdb_domain(sdb_domain, region)
                    print('SimpleDB Domain Deletions Successful.')

    if not options.skip_iam_instance_profiles:
        instance_profile_names = find_instance_profile_names_to_cleanup(options.include_all, match)
        if not options.view_only:
            if not instance_profile_names:
                print('Nothing to be done.')
            else:
                response = input(f'\nDo you wish to delete these IAM instance profiles names in account: {account_name}?  (Y)es (N)o: ')
                if response.lower() in ('y', 'yes'):
                    print('\nOkay, now deleting...')
                    for instance_profile_name, region in instance_profile_names.items():
                        delete_iam_instance_profile(instance_profile_name, region)
                    print('Instance Profile Deletions Successful.')

    if not options.skip_iam_roles:
        iam_roles = find_iam_roles_to_cleanup(options.include_all, match)
        if not options.view_only:
            if not iam_roles:
                print('Nothing to be done.')
            else:
                response = input(f'\nDo you wish to delete these IAM roles in account: {account_name}?  (Y)es (N)o: ')
                if response.lower() in ('y', 'yes'):
                    print('\nOkay, now deleting...')
                    for iam_role, region in iam_roles.items():
                        delete_iam_role(iam_role, region)
                    print('Role Deletions Successful.')


if __name__ == '__main__':
    main(sys.argv[1:])



# Modified from: https://github.com/common-workflow-language/workflow-service
import functools
import json
import logging
import os
from abc import abstractmethod
from typing import Any, Callable, Optional, Union
from urllib.parse import urldefrag

import connexion  # type: ignore
from werkzeug.utils import secure_filename

from toil.lib.io import mkdtemp

logger = logging.getLogger(__name__)

# Define a type for WES task log entries in responses
# TODO: make this a typed dict with all the WES task log field names and their types.
TaskLog = dict[str, Union[str, int, None]]


class VersionNotImplementedException(Exception):
    """
    Raised when the requested workflow version is not implemented.
    """

    def __init__(
        self,
        wf_type: str,
        version: Optional[str] = None,
        supported_versions: Optional[list[str]] = None,
    ) -> None:
        if version:
            message = (
                "workflow_type '{}' requires 'workflow_type_version' to be one of '{}'.  "
                "Got '{}' instead.".format(wf_type, str(supported_versions), version)
            )
        else:
            message = f"workflow_type '{wf_type}' is not supported."

        super().__init__(message)


class MalformedRequestException(Exception):
    """
    Raised when the request is malformed.
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)


class WorkflowNotFoundException(Exception):
    """
    Raised when the requested run ID is not found.
    """

    def __init__(self) -> None:
        super().__init__("The requested workflow run wasn't found.")


class WorkflowConflictException(Exception):
    """
    Raised when the requested workflow is not in the expected state.
    """

    def __init__(self, run_id: str):
        super().__init__(f"Workflow {run_id} exists when it shouldn't.")


class OperationForbidden(Exception):
    """
    Raised when the request is forbidden.
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)


class WorkflowExecutionException(Exception):
    """
    Raised when an internal error occurred during the execution of the workflow.
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)


def handle_errors(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    This decorator catches errors from the wrapped function and returns a JSON
    formatted error message with the appropriate status code defined by the
    GA4GH WES spec.
    """

    def error(msg: Any, code: int = 500) -> tuple[dict[str, Any], int]:
        logger.warning(
            f"Exception raised when calling '{func.__name__}()':", exc_info=True
        )
        return {"msg": str(msg), "status_code": code}, code

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except MalformedRequestException as e:
            return error(e, code=400)
        except VersionNotImplementedException as e:
            return error(e, code=400)
        except OperationForbidden as e:
            return error(e, code=403)
        except (FileNotFoundError, WorkflowNotFoundException) as e:
            return error(e, code=404)
        except WorkflowConflictException as e:
            return error(e, code=400)
        except WorkflowExecutionException as e:
            return error(e, code=500)
        except Exception as e:
            return error(e, code=500)

    return wrapper


class WESBackend:
    """
    A class to represent a GA4GH Workflow Execution Service (WES) API backend.
    Intended to be inherited. Subclasses should implement all abstract methods
    to handle user requests when they hit different endpoints.
    """

    def __init__(self, options: list[str]):
        """
        :param options: A list of default engine options to use when executing
                        a workflow.  Example options:
                        ["--logLevel=CRITICAL","--workDir=/path/to/dir",
                            "--tag=Name=default", "--tag=Owner=shared", ...]
        """
        self.options = options or []

    def resolve_operation_id(self, operation_id: str) -> Any:
        """
        Map an operationId defined in the OpenAPI or swagger yaml file to a
        function.

        :param operation_id: The operation ID defined in the specification.
        :returns: A function that should be called when the given endpoint is
                  reached.
        """
        return getattr(self, operation_id.split(".")[-1])

    @abstractmethod
    def get_service_info(self) -> dict[str, Any]:
        """
        Get information about the Workflow Execution Service.

        GET /service-info
        """
        raise NotImplementedError

    @abstractmethod
    def list_runs(
        self, page_size: Optional[int] = None, page_token: Optional[str] = None
    ) -> dict[str, Any]:
        """
        List the workflow runs.

        GET /runs
        """
        raise NotImplementedError

    @abstractmethod
    def run_workflow(self) -> dict[str, str]:
        """
        Run a workflow. This endpoint creates a new workflow run and returns
        a `RunId` to monitor its progress.

        POST /runs
        """
        raise NotImplementedError

    @abstractmethod
    def get_run_log(self, run_id: str) -> dict[str, Any]:
        """
        Get detailed info about a workflow run.

        GET /runs/{run_id}
        """
        raise NotImplementedError

    @abstractmethod
    def cancel_run(self, run_id: str) -> dict[str, str]:
        """
        Cancel a running workflow.

        POST /runs/{run_id}/cancel
        """
        raise NotImplementedError

    @abstractmethod
    def get_run_status(self, run_id: str) -> dict[str, str]:
        """
        Get quick status info about a workflow run, returning a simple result
        with the overall state of the workflow run.

        GET /runs/{run_id}/status
        """
        raise NotImplementedError

    @staticmethod
    def log_for_run(run_id: Optional[str], message: str) -> None:
        if run_id:
            logging.info("Workflow %s: %s", run_id, message)
        else:
            logging.info(message)

    @staticmethod
    def secure_path(path: str) -> str:
        return os.path.join(
            *[
                str(secure_filename(p))
                for p in path.split("/")
                if p not in ("", ".", "..")
            ]
        )

    def collect_attachments(
        self, run_id: Optional[str], temp_dir: Optional[str]
    ) -> tuple[str, dict[str, Any]]:
        """
        Collect attachments from the current request by staging uploaded files
        to temp_dir, and return the temp_dir and parsed body of the request.

        :param run_id: The run ID for logging.
        :param temp_dir: The directory where uploaded files should be staged.
                         If None, a temporary directory is created.
        """
        if not temp_dir:
            temp_dir = mkdtemp()
        body: dict[str, Any] = {}
        has_attachments = False
        for key, ls in connexion.request.files.lists():
            try:
                for value in ls:
                    # uploaded files that are required to execute the workflow
                    if key == "workflow_attachment":
                        # guard against maliciously constructed filenames
                        dest = os.path.join(temp_dir, self.secure_path(value.filename))
                        if not os.path.isdir(os.path.dirname(dest)):
                            os.makedirs(os.path.dirname(dest))
                        self.log_for_run(
                            run_id, f"Staging attachment '{value.filename}' to '{dest}'"
                        )
                        value.save(dest)
                        has_attachments = True
                        body[key] = (
                            f"file://{temp_dir}"  # Reference to temp working dir.
                        )

                    elif key in (
                        "workflow_params",
                        "tags",
                        "workflow_engine_parameters",
                    ):
                        content = value.read()
                        body[key] = json.loads(content.decode("utf-8"))
                    else:
                        body[key] = value.read().decode()
            except Exception as e:
                raise MalformedRequestException(f"Error reading parameter '{key}': {e}")

        for key, ls in connexion.request.form.lists():
            try:
                for value in ls:
                    if not value:
                        continue
                    if key in ("workflow_params", "tags", "workflow_engine_parameters"):
                        body[key] = json.loads(value)
                    else:
                        body[key] = value
            except Exception as e:
                raise MalformedRequestException(f"Error reading parameter '{key}': {e}")

        if "workflow_url" in body:
            url, ref = urldefrag(body["workflow_url"])
            if ":" not in url:
                if not has_attachments:
                    raise MalformedRequestException(
                        "Relative 'workflow_url' but missing 'workflow_attachment'"
                    )
                body["workflow_url"] = self.secure_path(url)  # keep this relative
                if ref:
                    # append "#ref" after the url
                    body["workflow_url"] += "#" + self.secure_path(ref)
            self.log_for_run(
                run_id, "Using workflow_url '%s'" % body.get("workflow_url")
            )
        else:
            raise MalformedRequestException("Missing 'workflow_url' in submission")

        if "workflow_params" in body and not isinstance(body["workflow_params"], dict):
            # They sent us something silly like "workflow_params": "5"
            raise MalformedRequestException(
                "Got a 'workflow_params' which does not decode to a JSON object"
            )

        return temp_dir, body
